#include <string.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "global.h"
#include "edit.h"
#include "icon.h"
#include "menu.h"
#include "obj.h"
#include "qed.h"
#include "rsc.h"
#include "scroll.h"
#include "text.h"
#include "windows.h"
#include "comm.h"


/******************************************************************************/
/*		lokale Variablen ********************************************************/
/******************************************************************************/
LOCAL WORD	msgbuf[8];

#define GLOB_SIZE 256
LOCAL UBYTE *global_str;

#if 0
LOCAL WORD find_server(VOID)
{
	WORD	i;
	UBYTE server[] = "		  ",
			*s;
		
	i = appl_find("GEMINI  ");
	if (i < 0)
	{
		s = getenv("AVSERVER");
		if (s != NULL)
		{
			i = 0;
			while (s[i] != '\0')
				server[i++] = s[i];
			i = appl_find(server);
		}
		else
			i = -1;
	}
	if (i >= 0) 
		return i;
	else
		return -1;
}
#endif


/* 
 * Generiert aus einem Pfadnamen den GEM-Namen (max. acht Zeichen, ggf.
 * mit ' ' aufgefllt) und berprft ihn auf ST-GUIDE.
*/
LOCAL BOOLEAN make_name(UBYTE *longname, UBYTE *shortname, BOOLEAN must_stg)
{
	WORD	len, i;
	UBYTE	*p;
	
	if (longname[0] == EOS)
	{
		note(1, NOHELP);
		strcpy(shortname, "");
		return FALSE;
	}
	else
	{
		file_name(longname, shortname);
		p = strrchr(shortname, '.');				/* Extension abschneiden */
		*p = '\0';						
		len = (short) strlen(shortname);
		if (len < 8)									/* mit ' ' auffllen */
		{
			for (i = len; i < 8; i++)
				strcat(shortname, " ");
		}
		if (must_stg && (strcmp(shortname, "ST-GUIDE") != 0))
		{
			note(1, NOSTG);
			return FALSE;
		}
		else
			return TRUE;
	}
}


/*
 * Liefert die app_id des bergebenen Programmes zurck. Falls es noch
 * nicht luft, wird es mit arg gestartet.
*/
LOCAL WORD get_id(UBYTE* name, UBYTE *path, UBYTE *arg, BOOLEAN *started)
{
	WORD		ret = -1;
	UBYTE		*p, help[128] = "";
	BOOLEAN	s = FALSE;
		
	if (name[0] != EOS)
	{
		ret = appl_find(name);
		if (ret < 0)							/* ggf. als Programm nachstarten */
		{
			if (arg != NULL)
			{
				strcpy(help, " ");
				strcat(help, arg);
				help[0] = (UBYTE)strlen(help);
			}
			p = strrchr(path, '.');
			if (file_exist(path) && ((strncmp(p, ".APP", 3) == 0) || (strncmp(p, ".PRG", 3) == 0)))
			{
				if (mtos)
				{
					ret = shel_write(1, 1, 1, path, help);
					s = (ret > 0);
				}
				else if (magx)
				{
					ret = shel_write(1, 1, 100, path, help);
					s = (ret > 0);
				}
				else
				{
					note(1, HELPSTART);
					ret = -1;
				}
			}
			else
			{
				snote(1, HELPNOTF, helpprog);	
				ret = - 1;
			}
		}
	}
	if (started != NULL)
		*started = s;
	return ret;
}

/******************************************************************************/
/* 	VA-Protokol 																				*/
/******************************************************************************/

#define AC_HELP		1025


GLOBAL VOID send_vastart(WORD id, CONST UBYTE *data)
{
	strcpy(global_str, data);
	msgbuf[0] = VA_START;
	msgbuf[1] = gl_apid;
	msgbuf[2] = 0;
	*(UBYTE **)(msgbuf + 3) = global_str;
	msgbuf[5] = 0;
	msgbuf[6] = 0;
	msgbuf[7] = 0;
	appl_write(id, 16, msgbuf);
}


LOCAL VOID send_help(WORD id, CONST UBYTE *str)
{
	strcpy(global_str, str);
	msgbuf[0] = AC_HELP; 					/* magic message number */
	msgbuf[1] = gl_apid; 					/* my own id */
	msgbuf[2] = 0; 							/* no more than 16 UBYTEs */
	*(UBYTE**)&msgbuf[3] = global_str;	/* the KeyWord */
	appl_write(id, 16, msgbuf);			/* write message */
}


GLOBAL BOOLEAN call_help(CONST UBYTE *str)
{
	WORD		help_id;
	FILENAME	name;
			
	Wake_mouse();
	if (make_name(helpprog, name, FALSE))
	{
		help_id = get_id(name, helpprog, NULL, NULL);
		if (help_id > 0)
		{
			send_help(help_id, str);
			return TRUE;
		}
	}
	return FALSE;
}


GLOBAL BOOLEAN call_hyp(CONST UBYTE *data)
{
	WORD		stg_id;
	UBYTE 	help[256];
	FILENAME name;
	BOOLEAN	started;
	
	if (helpprog[0] == EOS)
	{
		note(1, NOHELP);
		return FALSE;
	}
	Wake_mouse();
	if (make_name(helpprog, name, TRUE))
	{
		strcpy(help, "*:\\QED.HYP ");
		strcat(help, data);
		stg_id = get_id(name, helpprog, help, &started);
		if (stg_id > 0 && !started)			/* nicht gestartet, da es bereits lief! */
			send_vastart(stg_id, help);		/* -> also Message schicken */
		return (stg_id > 0);
	}
	return FALSE;
}


/******************************************************************************/
/* 	SE-Protokoll																				*/
/******************************************************************************/

/*
 * exportierte Variablen
 */
GLOBAL ShellEntry SE_Shells[4];			/* 0 - 2 : QED sucht Shell */
													/* 3		: Shell sucht QED */

GLOBAL PATH 		SE_Makefile;
GLOBAL BOOLEAN 	SE_Activ;				/* haben wir Kontakt zur Shell? */


typedef struct _separm
{
		WORD		id;
		WORD		shellCmd;
		WORD		editCmd;
} SEPARMS;


#define WAITTIME	2000				/* Zeit, die QED auf ein SE_ACK wartet */

LOCAL SEPARMS	shellParm,
					editParm;
LOCAL WORD		msgbuff[8]; 		/* Buffer zum Senden! */
LOCAL WORD		aktiv;				/* Nummer der aktiven Shell (0..3) */
LOCAL WORD		menu_len;

/* lokale Prototypen */
LOCAL BOOLEAN	waitAck(VOID);
LOCAL VOID		sendOK(VOID);
LOCAL VOID		sendACK(BOOLEAN ok);
LOCAL VOID		sendCmd(WORD cmd, WORD bit);
LOCAL BOOLEAN	sendMsg(WORD id);


LOCAL BOOLEAN	sendMsg(WORD id)
{
	WORD	ret;
	
	ret = appl_write(id, (short) sizeof(msgbuff), msgbuff);
	return (ret > 0);
}

LOCAL VOID	setupMakefile(VOID)
{
	FILENAME n = "";

	if (SE_Makefile[0] == EOS)
		strcpy(n, strings(SEMAKESTR));		
	else
		file_name(SE_Makefile, n);
	fillup_menu(MSMAKEFILE, n, 2);
}


LOCAL VOID setTitel(UBYTE *titel)
{
	UBYTE str[11] = " ";

	strcat(str, titel);
	strcat(str, " ");
	menu_bar(menu, 0);
	menu[2].ob_width = menu_len + (short)strlen(str) * sys_wchar;
	menu[MSHELL].ob_width = (short)strlen(str) * sys_wchar;
	strcpy(menu[MSHELL].ob_spec.free_string, str);
	menu_bar(menu, 1);
}

GLOBAL VOID setupShellMenu(BOOLEAN text_win)
{
	WORD	i;
	
	if (SE_Activ)
	{
		setupMakefile();
		set_menu(MSMAKEFILE, (shellParm.editCmd & _ESMAKE));
		set_menu(MSCOMP, ((shellParm.editCmd & _ESCOMPILE) && text_win));
		set_menu(MSMAKE, (shellParm.editCmd & _ESMAKE));
		set_menu(MSMAKEALL, (shellParm.editCmd & _ESMAKEALL));
		set_menu(MSLINK, ((shellParm.editCmd & _ESLINK) && text_win));
		set_menu(MSEXEC, ((shellParm.editCmd & _ESEXEC) && text_win));
		set_menu(MSMAKEEXEC, (shellParm.editCmd & _ESMAKEEXEC));
	}
	else
	{													/* SE beendet, alles aus */
		for (i = MSMAKEFILE; i <= MSMAKEEXEC; i++)
			set_menu(i, FALSE);
	}
}


LOCAL BOOLEAN	waitAck(VOID)
{
	WORD	msg[8], d, ev;

	while (TRUE)
	{
		ev = evnt_multi ((MU_MESAG | MU_TIMER),
								0, 0, 0, 0, 0, 0, 0, 0,
								0, 0, 0, 0, 0,
								msg, WAITTIME, 0,
								&d, &d, &d, &d, &d, &d);
		if (ev == MU_TIMER)
		{
			SE_Activ = FALSE;
			setTitel("Shell");
			updt_menu(NULL);
			return FALSE;
		}
		else if (ev == MU_MESAG)
			if (msg[0] == SE_ACK)
			{
				return TRUE;
			}
	}		
}


LOCAL BOOLEAN	waitSEOK(VOID)
{
	WORD	msg[8], d, ev;

	while (TRUE)
	{
		ev = evnt_multi ((MU_MESAG | MU_TIMER),
								0, 0, 0, 0, 0, 0, 0, 0,
								0, 0, 0, 0, 0,
								msg, WAITTIME, 0,
								&d, &d, &d, &d, &d, &d);
		if (ev == MU_TIMER)
			return FALSE;
		else if (ev == MU_MESAG)
			if (msg[0] == SE_OK)
			{
				shellParm.id		 = msg[1];
				shellParm.shellCmd = msg[3];
				shellParm.editCmd  = msg[5];
				SE_Activ = TRUE;
				setTitel(SE_Shells[aktiv].Name);
				updt_menu(NULL);
				return TRUE;
			}
	}		
}

/* Shell -> Editor ************************************************************/

/* Antwort auf SE_INIT */				
LOCAL VOID	sendOK(VOID)
{
	if (shellParm.editCmd & _ESOK)
	{
		msgbuff[0] = ES_OK;
		msgbuff[1] = editParm.id;
		msgbuff[2] = 0;
		msgbuff[3] = editParm.shellCmd;
		msgbuff[5] = editParm.editCmd;
		msgbuff[6] = SEPROTOVERSION;
		msgbuff[7] = shellParm.id;
		sendMsg(shellParm.id);
	}
}


LOCAL VOID	sendACK(BOOLEAN ok)
{
	if (shellParm.editCmd & _ESACK)
	{
		msgbuff[0] = ES_ACK;
		msgbuff[1] = editParm.id;
		msgbuff[2] = 0;
		msgbuff[3] = ok;
		sendMsg(shellParm.id);
	}
}


GLOBAL VOID hndl_SEprotokol(WORD *msg)
{
	UBYTE 	datei[255], fehler[255], str[9];
	WORD		icon;
	ERRINFO	*error;
	
	switch (msg[0])
	{
		case SE_INIT :
			shellParm.id = msg[1];
			shellParm.shellCmd = msg[3];
			shellParm.editCmd  = msg[5];
			sendOK();
			SE_Activ = TRUE;
			aktiv = 3;

			/* Wir versuchen den GEM-Namen des Senders zu ermitteln */
			if (magx)
			{
				sprintf(str, "? %cXXXXX", (char) shellParm.id);
				str[1] = '\0';
				appl_find(str);
				strcpy(SE_Shells[aktiv].Name, str);
			}
			else if (mtos)
			{
				WORD	type, id, i;

				i = appl_search( 0, str, &type, &id);
				while (i != 0)
				{ 
					if (type == 2 && id == shellParm.id)
						break;
					i = appl_search( 1, str, &type, &id); 
				}
				strcpy(SE_Shells[aktiv].Name, str);
			}
			else
				strcpy(str, "Shell");
			setTitel(str);
			updt_menu(NULL);
			break;
		case SE_OPEN :
			strcpy(datei, *(UBYTE **)(msg + 3));
			sendACK(TRUE);
			icon = load_edit(datei, FALSE, -1, -1, NULL);
			if (icon > 0)
				do_icon(icon, DO_OPEN);
			break;
		case SE_ERROR :
			error = *(ERRINFO **)(msg + 3);
			strcpy(datei, error->errFile);
			strcpy(fehler, error->errMess);
			desire_x = error->errRow - 1;
			desire_y = error->errLine - 1;
			sendACK(TRUE);
			icon = load_edit(datei, FALSE, -1, -1, NULL);
			if (icon > 0)
			{
				TEXTP t_ptr;

				t_ptr = get_text(icon);
				strcpy(t_ptr->m2_err_str, fehler);
				do_icon(icon, DO_OPEN);
				Icon_edit(icon, DO_GOTO);
			}
			break;
		case SE_ERRFILE :
			strcpy(datei, *(UBYTE **)(msg + 5));
			strcpy(fehler, *(UBYTE **)(msg + 3));	/* Fehlerdatei */
			sendACK(TRUE);
			icon = load_edit(datei, FALSE, -1, -1, NULL);
			if (icon > 0)
				do_icon(icon, DO_OPEN);
			icon = load_edit(fehler, FALSE, -1, -1, NULL);
			if (icon > 0)
				do_icon(icon, DO_ABAND);
			break;
		case SE_QUIT :
			SE_Activ = FALSE;
			setTitel("Shell");
			updt_menu(NULL);
			note(1, SECLOSE);
			break;
		case SE_ACK :
			/* Shell hat besttigt. Toll. */
			break;
		default:
			inote(1, SEUNKNOWN, msg[0]);
			break;
	} /* switch */
}


/* Editor -> Shell ************************************************************/


LOCAL VOID	sendESQUIT(VOID)
{
	msgbuff[0] = ES_QUIT;
	msgbuff[1] = editParm.id;
	msgbuff[2] = 0;
	(VOID) sendMsg(shellParm.id);
	SE_Activ = FALSE;
}


LOCAL VOID	sendCmd(WORD cmd, WORD bit)
{
	if (shellParm.editCmd & bit)
	{
		msgbuff[0] = cmd;
		msgbuff[1] = editParm.id;
		msgbuff[2] = 0;
		if (global_str[0] != EOS)
			*(UBYTE **) &msgbuff[3] = global_str;
		else
			*(UBYTE **) &msgbuff[3] = NULL;
		if (!sendMsg(shellParm.id))
		{
			SE_Activ = FALSE;
			setTitel("Shell");
			updt_menu(NULL);
			note(1, SENOANS);
			return;
		}
		if (!waitAck())
			note(1, SENOANS);
	}
}


LOCAL BOOLEAN	get_filename(VOID)
{
	WINDOWP	window;
	TEXTP 	t_ptr;

	window = top();
	if (window != NULL && window->class == CLASS_EDIT)
	{
		t_ptr = get_text(window->link);
		strcpy(global_str, t_ptr->filename);
		return TRUE;
	}
	else
		return FALSE;
}

LOCAL VOID	suchShell(VOID)
{
	WORD	i, antw, len;
	UBYTE name[9];
	
	for (i = 0; i < 3; i++) 
	{
		set_ptext(suchshell, SENAME1 + i, SE_Shells[i].Name);
		set_state(suchshell, SEUSE1 + i, SE_Shells[i].aktiv);
	}
	Arrow_mouse();
	antw = HndlDial(suchshell, SENAME1, FALSE, NULL, NULL);
	Last_mouse();
	if (antw == SSOK)
	{
		if (SE_Activ)
		{
			antw = note(2, SEONLINE);
			if (antw == 2)
				return;
			else
			{
				sendESQUIT();
				setTitel("Shell");
			}
		}
		
		for (i = 0; i < 3; i++) 
		{
			get_ptext(suchshell, SENAME1 + i, SE_Shells[i].Name);
			SE_Shells[i].aktiv = get_state(suchshell, SEUSE1 + i);
			if (SE_Shells[i].aktiv)
			{
				aktiv = i;
				strcpy(name, SE_Shells[aktiv].Name);
			}
		}
		
		if (name[0] != EOS)
		{
				len = (short) strlen(name);
				if (len < 8)					/* mit ' ' auffllen */
					for (i = len; i < 8; i++)
						strcat(name, " ");
				i = appl_find(name);
				if (i < 0) 
				{
					note(1, SENOTFOUND);
					SE_Activ = FALSE;
					setTitel("Shell");
					updt_menu(NULL);
					return;
				}
				msgbuff[0] = ES_INIT;
				msgbuff[1] = editParm.id;
				msgbuff[2] = 0;
				msgbuff[3] = editParm.shellCmd;
				msgbuff[4] = 0;
				msgbuff[5] = editParm.editCmd;
				sendMsg(i);
				if (!waitSEOK())
					note(1, SENOANS);
		}
		else
					note(1, SENONAME);
	}
}


GLOBAL VOID hndl_ESprotokol(WORD item)
{
	FSEL		fsel = {"", "*.*"};
	PATH		name = "";
	FILENAME n = "";


	if (item == MSSUCH)
		suchShell();

	if (!SE_Activ)
		return ;

	switch (item)
	{
		case MSMAKEFILE:
			if (global_shift && SE_Makefile[0] != EOS)	/* Makefile lschen */
			{
				SE_Makefile[0] = EOS;
				updt_menu(NULL);
				break;
			}
			if (SE_Makefile[0] != EOS)
			{
				file_name(SE_Makefile, n);
				set_fsel_path(SE_Makefile);
			}
			strcpy(fsel.name, n);
			if (select_file(&fsel, name, strings(FINDMKSTR)))
			{
				strcpy(SE_Makefile, name);
				setupMakefile();
				sendCmd(ES_PROJECT, _ESPROJECT);
			}
			break;
		case MSCOMP :
			if (get_filename())
				sendCmd(ES_COMPILE, _ESCOMPILE);
			break;
		case MSMAKE :
			strcpy(global_str, SE_Makefile);
			sendCmd(ES_MAKE, _ESMAKE);
			break;
		case MSMAKEALL :
			strcpy(global_str, SE_Makefile);
			sendCmd(ES_MAKEALL, _ESMAKEALL);
			break;
		case MSLINK :
			if (get_filename())
				sendCmd(ES_LINK, _ESLINK);
			break;
		case MSEXEC :
			if (get_filename())
				sendCmd(ES_EXEC, _ESEXEC);
			break;
		case MSMAKEEXEC :
			strcpy(global_str, SE_Makefile);
			sendCmd(ES_MAKEEXEC, _ESMAKEEXEC);
			break;
	} /* switch */
}


LOCAL VOID	init_seprotokol(VOID)
{
	SE_Activ = FALSE;
	menu_len = menu[2].ob_width - menu[MSHELL].ob_width;
	setTitel("Shell");
	
	shellParm.id		 = -1;
	shellParm.shellCmd = 0;
	shellParm.editCmd  = 0;
		
	editParm.id 		 = gl_apid;
	editParm.shellCmd  = (_SEINIT|_SEOK|_SEACK|_SEQUIT|_SEOPEN|_SEERROR|_SEERRFILE);
	editParm.editCmd	 = (_ESINIT|_ESOK|_ESACK|_ESQUIT|_ESCOMPILE|_ESMAKE|_ESMAKEALL|_ESLINK|_ESEXEC|_ESMAKEEXEC|_ESPROJECT);

	setupMakefile();
}


/*****************************************************************************/
/* 	Drag&Drop																				  */
/*****************************************************************************/
#define DD_OK			0
#define DD_NAK 		1
#define DD_EXT 		2
#define DD_LEN 		3

#define DD_NUMEXTS	8
#define DD_EXTSIZE	32
#define DD_NAMEMAX	128

LOCAL VOID	*oldsig;

LOCAL WORD parse_ARGS(UBYTE *str)
{
	WORD		cnt = 1;
	UBYTE 	*c = str;
	BOOLEAN	in_quote = FALSE;
	
	while (*c)
	{
		switch (*c)
		{
			case ' ' :
				if (!in_quote)
				{
					*c = '\0';
					cnt++;
				}
				break;
			case 39 :	/* ' */
				strcpy(c, c + 1);
				if (!in_quote)
					in_quote = TRUE;
				else
				{
					if (*c != 39)
					{
						in_quote = FALSE;
						*c = 0;
						if (c[1])
							cnt++;
					}
				}
				break;
			default:
				break;
		} /* switch */
		c += 1;
	} /* while */
	return cnt;
}

LOCAL VOID insert(WINDP w, WORD kstate, UBYTE *filename)
{
	global_shift = (kstate & 3);
	strcpy(drag_filename, filename);

	/*
	 * Falls ein Pfad gedragged wurde, setzen wir global_shift, damit 
	 * er ggf. im Textfenster als Text eingesetzt wird!!
	 * Wird ein Pfad auf ein Projekt gezogen, hat global_shift kein
	 * Einflu, sondern es wird die Dateisuche mit dem Pfad als Suchpfad
	 * gestartet!
	 */
	if (filename[strlen(filename) - 1] == '\\' || global_shift)
		Icon_drag(w->link, DRAGDROP_PATH);
	else
		Icon_drag(w->link, DRAGDROP_FILE);
}

LOCAL VOID parse(UBYTE *obname, UBYTE *cmdline, WINDP w, WORD kstate)
{
	WORD	comps = parse_ARGS(cmdline);
	UBYTE *c = cmdline;
	
	while (comps - 1)
	{
		insert(w, kstate, c);
		c += strlen(c) +1;
		comps--;
	}
	insert(w, kstate, c);	
}

LOCAL LONG dd_open_pipe(WORD pnum)
{
	UBYTE pipename[20];
	LONG	fd;
	
	sprintf(pipename, "U:\\PIPE\\DRAGDROP.%c%c", (pnum & 0xff00) >> 8, pnum & 0xff);
	if ((fd = Fopen(pipename, 2)) > 0)
		oldsig = Psignal(SIGPIPE, (void *) 1);
	return fd;
}

LOCAL VOID dd_close(LONG fd)
{
	Psignal(SIGPIPE, oldsig);
	Fclose((short) fd);
}

LOCAL LONG dd_open(WORD pipe_num, CONST UBYTE *extlist)
{
	LONG	fd;
	UBYTE outbuf[DD_EXTSIZE + 1];
	
	fd = dd_open_pipe(pipe_num);
	if (fd < 0)
		return 0L;
		
	outbuf[0] = DD_OK;
	strncpy(outbuf + 1, extlist, DD_EXTSIZE);
	if (Fwrite((short) fd, DD_EXTSIZE + 1, outbuf) != DD_EXTSIZE + 1)
	{
		dd_close(fd);
		return -1L;
	}
	return fd;
}

LOCAL BOOLEAN dd_getheader(LONG fd, UBYTE *obname, UBYTE *fname, UBYTE *datatype,
									LONG *size)
{
	WORD		hdrlen, cnt, slen;
	UBYTE 	buf[PATH_MAX + DD_NAMEMAX + 1];
	
	if (Fread((short) fd, 2, &hdrlen) != 2)
		return FALSE;
	if (hdrlen < 9)
		return FALSE;
	
	if (Fread((short) fd, 4, datatype) != 4)
		return FALSE;
	datatype[4] = '\0';
	
	if (Fread((short) fd, 4, size) != 4)
		return FALSE;
	hdrlen -= 8;
	cnt = hdrlen;
	
	if (cnt > PATH_MAX + DD_NAMEMAX)
		cnt = PATH_MAX + DD_NAMEMAX;
	if (Fread((short) fd, cnt, buf) != cnt)
		return FALSE;
	buf[PATH_MAX + DD_NAMEMAX] = '\0';
	hdrlen -= cnt;
	slen = (short) strlen(buf);
	
	if (slen < DD_NAMEMAX)
		strcpy(obname, buf);
	if (slen < PATH_MAX + DD_NAMEMAX)
	{
		UBYTE *fp = buf + slen + 1;
		
		slen = (short) strlen(fp);
		if (slen < PATH_MAX)
			strcpy(fname, fp);
	}
	
	while (hdrlen)
	{
		WORD	cnt = hdrlen;
		
		if (cnt > sizeof(buf))
			hdrlen = (short) sizeof(buf);
		Fread((short) fd, cnt, buf);
		hdrlen -= cnt;
	}
	return TRUE;
}

LOCAL BOOLEAN dd_reply(LONG fd, UBYTE ack)
{
	if (Fwrite((short) fd, 1L, &ack) != 1)
	{
		dd_close(fd);
		return FALSE;
	}
	return TRUE;
}

VOID	hndl_DD(WORD *msg)
{
	WINDP w_dest = find_window(msg[3]);
	LONG	fd,
			size;
	UBYTE obname[DD_NAMEMAX],
			ext[5],
			fname[PATH_MAX],
			ourexts[DD_EXTSIZE] = "ARGS.TXT.ASC";

	if (w_dest != NULL)
	{
		fd = dd_open(msg[7], ourexts);
		if (fd < 0)
			return ;
	
		do
		{
			if (!dd_getheader(fd, obname, fname, ext, &size))
			{
				dd_close(fd);
				return;
			}
			if (!strncmp(ext, "ARGS", 4))
			{
				UBYTE	*cmdline = malloc(size +1);
				
				if (!cmdline)
				{
					dd_reply(fd, DD_LEN);
					continue;
				}
				dd_reply(fd, DD_OK);
				Fread((short) fd, size, cmdline);
				dd_close(fd);
				cmdline[size] = 0;
				parse(obname, cmdline, w_dest, msg[6]);
				free(cmdline);
				return ;
			}
			if (strncmp(ext, ".TXT", 4) == 0 ||
				 strncmp(ext, ".ASC", 4) == 0 )
			{
				drag_data = Malloc(size + 1);
				/* ^^ wird nach dem Einfgen in edit.c wieder freigegeben! */
				drag_data_size = size;
				if (drag_data == NULL)
				{
					dd_reply(fd, DD_LEN);
					continue;
				}
				dd_reply(fd, DD_OK);
				Fread((short) fd, drag_data_size, drag_data);
				dd_close(fd);
				drag_data[drag_data_size] = 0;
				Icon_drag(w_dest->link, DRAGDROP_DATA);
				return ;
			}
		}
		while (dd_reply(fd, DD_EXT));
	}
}

/*****************************************************************************/
/* 	Sonstiges																				  */
/*****************************************************************************/

GLOBAL VOID init_comm(VOID)
{
	LONG	ret;
	WORD	mxMask;
	
	init_seprotokol();

	mxMask = Mxmask();
	if (mxMask)
		ret = (LONG) Mxalloc(GLOB_SIZE, 0x43 & mxMask);
	else	
		ret = (LONG) Malloc(GLOB_SIZE);

	if (ret > 0L)
		global_str = (UBYTE *)ret;
}

GLOBAL VOID term_comm(VOID)
{
	if (SE_Activ && (shellParm.editCmd & _ESQUIT))
		sendESQUIT();
	if (global_str != NULL)
		Mfree(global_str);
}
