#include <string.h>		/* str.. */
#include <stdlib.h>		/* itoa, atoi, getenv */
#include <stdarg.h>		/* va_.. */

#include "global.h"
#include "ausgabe.h"		/* BildPos */
#include "clipbrd.h"		/* (is_)clip_ondisk */
#include "comm.h"
#include "desktop.h"		/* add_icon_to_desk */
#include "disk.h"			/* add_diskinfo */
#include "edit.h"			/* font_edit */
#include "error.h"
#include "find.h"
#include "fontsel.h"		/* winFont, druckFont */
#include "gdos.h"
#include "icon.h"			/* add_icon */
#include "kurzel.h"		/* get/set_marke */
#include "makro.h"		/* set_makro_str, get_makro_str */
#include "menu.h"
#include "printer.h"		/* printer_type */
#include "projekt.h"
#include "qed.h"
#include "rsc.h"
#include "scroll.h"		/* COPYB */
#include "set.h"
#include "text.h"
#include "trash.h"		/* trash_type */
#include "umbruch.h"		/* set_umbruch */
#include "windows.h"
#include "file.h"

/*=================================================================*/

#define	INFVERSION	105		/* Version der DEFAULT.QED */


typedef struct
{
	WORD		handle;		/* File-Handle */
	LONG		left;			/* Rest im Puffer */
	LONG		size;			/* Gre des Puffers */
	UBYTE		*buffer;		/* Anfang des Puffers */
	UBYTE		*ptr;			/* Zeiger auf erstes freie Byte */
	BOOLEAN	cr_lf;		/* CR+LF oder nur LF */
	BOOLEAN	var;			/* Puffer ber Malloc geholt? */
	WORD		err;
}BUFFER, *BUFFER_P;

#define puff_size 		(4*1024L)
#define max_puff_size	(16*1024L)
#define rest_mem			(50*1024L)

LOCAL UBYTE	puffer[puff_size+2];		/* +2 wg. READ */
LOCAL FSEL	qed_fsel = {"DEFAULT.QED","*.QED"};

/*=================================================================*/

LOCAL VOID	back_up			(TEXTP t_ptr);
LOCAL VOID	restore_back_up(TEXTP t_ptr);
LOCAL VOID	kill_back_up	(TEXTP t_ptr);
LOCAL WORD	save_datei		(const UBYTE *name, RINGP t, BOOLEAN verbose, BOOLEAN cr_lf);
WORD			load_datei		(const UBYTE *name, RINGP t, BOOLEAN verbose, BOOLEAN *cr_lf);
LOCAL VOID 	WRITE				(BUFFER_P buffer, RINGP t, BOOLEAN verbose);
LOCAL WORD	READ				(WORD handle, RINGP t, BOOLEAN verbose, BOOLEAN *cr_lf);

/*=================================================================*/

LOCAL VOID init_buffer(BUFFER_P buffer, const UBYTE *name, BOOLEAN cr_lf, LONG anz)
{
	LONG	bytes;

	buffer->cr_lf = cr_lf;
	buffer->err = (short) Fcreate(name,0);
	if (buffer->err<0) return;

	buffer->handle = buffer->err;
	buffer->err = 0;
	bytes = (long)Malloc(-1L);
	bytes &= (~1023L);
	if (bytes > puff_size + rest_mem)			/* min. puff_size */
	{
		bytes -= rest_mem;
		anz = (anz>>3);
		anz &= (~1023L);
		anz = max(max_puff_size,anz);
		bytes = min(anz,bytes);
		buffer->buffer = Malloc(bytes);
		if (buffer->buffer!=NULL)
		{
			buffer->left = buffer->size = bytes;
			buffer->var = TRUE;
		}
		else
		{
			buffer->buffer = puffer;
			buffer->left = buffer->size = puff_size;
			buffer->var = FALSE;
		}
	}
	else
	{
		buffer->buffer = puffer;
		buffer->left = buffer->size = puff_size;
		buffer->var = FALSE;
	}
	buffer->ptr = buffer->buffer;
}

LOCAL VOID kill_buffer(BUFFER_P buffer)
{
	LONG	anz = buffer->size-buffer->left;

	if (anz && !buffer->err)
	{
		if (Fwrite(buffer->handle,anz,buffer->buffer)<anz)
			buffer->err = -39;
		buffer->left = buffer->size;
	}
	if (buffer->var)
		Mfree(buffer->buffer);
	Fclose(buffer->handle);
}

LOCAL VOID _write(BUFFER_P buffer, UBYTE *what, WORD len, BOOLEAN lastline)
{
	UBYTE	*ptr;
	LONG	bsize = buffer->size,
			bleft = buffer->left;

	if (buffer->err) 
		return;
	ptr = buffer->ptr;
	if (len >= bleft)
	{
		COPYB(ptr, what, (short) bleft);
		if (Fwrite(buffer->handle,bsize,buffer->buffer) < bsize)
		{
			buffer->left = 0;
			buffer->err = -39;
			return;
		}
		what += bleft;
		len -= (short) bleft;					/* Rest im String */
		ptr = COPYB(buffer->buffer,what,len);
		bleft = bsize;
	}
	else
	{
#if 0				/* MK@UN : Workaround, strtzt auf 68000 manchmal ab :-(((*/
		if ((bleft & 1) == 0)
		{
			COPYW(ptr, what, len);
			ptr += len;
		}
		else
#endif
			ptr = COPYB(ptr,what,len);
	}
	bleft -= len;
	if (buffer->cr_lf)
	{
		if (bleft < 2)
		{
			if (!lastline)
				*ptr = '\r';
			if (Fwrite(buffer->handle,bsize,buffer->buffer)<bsize)
			{
				buffer->left = 0;
				buffer->err = -39;
				return;
			}
			ptr = buffer->buffer;
			if (!lastline)
			{
				*ptr++ = '\n';
				bleft = bsize - 1;
			}
		}
		else
		{
			if (!lastline)
			{
				*ptr++ = '\r';
				*ptr++ = '\n';
				bleft -= 2;
			}
		}
	}
	else
	{
		if (!lastline)
		{
			*ptr++ = '\n';
			bleft -= 1;
		}
	}
	buffer->ptr = ptr;
	buffer->left = bleft;
	return;
}

LOCAL VOID restore_back_up(TEXTP t_ptr)
{
	if (t_ptr->loc_opt->backup)
	{
		PATH old;

		if (file_exist(t_ptr->filename))
			Fdelete(t_ptr->filename);
		strcpy(old, t_ptr->filename);
		make_ext(old, t_ptr->loc_opt->backup_ext);
		Frename(0, old, t_ptr->filename);
	}
}

LOCAL VOID kill_back_up(TEXTP t_ptr)
{
	PATH new;

	strcpy(new, t_ptr->filename);
	make_ext(new, t_ptr->loc_opt->backup_ext);
	if (file_exist(new))			/* DUP-Datei lschen */
		Fdelete(new);
}

LOCAL VOID back_up(TEXTP t_ptr)
{
	PATH new;

	if (t_ptr->loc_opt->backup)
	{
		Busy_mouse();
		strcpy(new, t_ptr->filename);
		make_ext(new, t_ptr->loc_opt->backup_ext);
		if (file_exist(new))			/* Alte DUP-Datei lschen */
			Fdelete(new);
		Frename(0, t_ptr->filename, new);
		Last_mouse();
	}
}

WORD save (TEXTP t_ptr)
{
	WORD	antw;

	if (file_exist(t_ptr->filename))
	{
		if (file_readonly(t_ptr->filename))
		{
			FILENAME	file;
			
			t_ptr->readonly = TRUE;
			file_name(t_ptr->filename, file);
			snote(1,READONLY, file);
			return -39;
		}
		else
			t_ptr->readonly = FALSE;
		if (t_ptr->file_date_time!=-1L)
		{
			LONG date_time = file_time(t_ptr->filename,NULL,NULL);
			
			if (date_time!=t_ptr->file_date_time)
			{
				FILENAME	name;
				
				file_name(t_ptr->filename, name);
				if (snote(1,MOVED2,name)==2) return -1;
			}
		}
	}
	back_up(t_ptr);
	antw = save_datei(t_ptr->filename,&t_ptr->text,TRUE,t_ptr->cr_lf);
	if (antw==0)
	{
		t_ptr->moved = 0;
		t_ptr->file_date_time = file_time(t_ptr->filename,NULL,NULL);
		t_ptr->count = local_time->tm_min;
	}
	else
	{
		if (antw==-39)
			note(1,NOSPACE);
		else
			note(1,WRITEERR);
		restore_back_up(t_ptr);
		t_ptr->file_date_time = -1L;
	}
	return(antw);
}

WORD save_as (TEXTP t_ptr, const UBYTE *name)
{
	WORD	antw;

	if (file_exist(name))
	{
		if (file_readonly(name))
		{
			FILENAME	file;
			
			file_name(t_ptr->filename, file);
			snote(1,READONLY, file);
			return -39;
		}
	}
	back_up(t_ptr);
	antw = save_datei(name,&t_ptr->text,TRUE,t_ptr->cr_lf);
	if (antw!=0)
	{
		if (antw==-39)
			note(1,NOSPACE);
		else
			note(1,WRITEERR);
		restore_back_up(t_ptr);
	}
	return(antw);
}

BOOLEAN save_new(UBYTE *name, FSEL *fsel, const UBYTE *text)
{
	fsel->name[0] = EOS;
	if (!select_file(fsel,name,text))
		return FALSE;
	if (file_exist(name))
	{
		if (note(1,EXIST)==2) return FALSE;
	}
	return TRUE;
}

LOCAL WORD save_datei(const UBYTE *name, RINGP t, BOOLEAN verbose, BOOLEAN cr_lf)
{
	BUFFER	buffer;
	LONG		bytes;

	Busy_mouse();
	bytes = textring_bytes(t, cr_lf);
	init_buffer(&buffer, name, cr_lf, bytes);
	if (buffer.err)
	{
		Last_mouse();
		return(buffer.err);
	}
	else
	{
		verbose = verbose && ((bytes>>10)>=transfer_size);
		if (verbose)
		{
			UBYTE	str[40];
			FILENAME	file;

			strcpy(str, strings(SAVESTR));
			file_name(name,file);
			strcat(str,file);
			start_aktion(str,FALSE,t->lines);
		}
		WRITE(&buffer, t, verbose);
		kill_buffer(&buffer);
		if (buffer.err < 0)	
			Fdelete(name);
		if (verbose) 
			end_aktion();
		Last_mouse();
		return(buffer.err);
	}
}

LOCAL VOID WRITE(BUFFER_P buffer, RINGP t, BOOLEAN verbose)
{
	LINEP start;
	WORD	ak, delta;
	LONG	lines = t->lines;

	start = t->head.nachf;
	ak = delta = (short) max((lines >> 6), 1);
	while ((--lines) >= 0 && !buffer->err)
	{
		if (verbose && !(--ak))
		{
			do_aktion(t->lines-lines);
			ak = delta;
		}
		_write(buffer, TEXT(start), start->len, (lines == 0));
		NEXT(start);
	}
	if (verbose) do_aktion(t->lines);		/* wg. der Schnheit */
}

/*==================================================================*/

WORD load(TEXTP t_ptr, BOOLEAN quiet)
{
	WORD	antw;

	antw = load_datei(t_ptr->filename, &t_ptr->text,TRUE, &t_ptr->cr_lf);
	t_ptr->cursor_line = t_ptr->text.head.nachf;
	t_ptr->readonly = file_readonly(t_ptr->filename);
	if (antw==0)
	{
		t_ptr->file_date_time = file_time(t_ptr->filename,NULL,NULL);
	}
	else
		t_ptr->file_date_time = -1L;
	if (antw!=0 && !quiet)
	{
		FILENAME	datei;
		PATH	path;
		
		file_splitt(t_ptr->filename, path, datei);
		
		if (antw==-33 && path_exist(path))	/* file not found und Pfad existiert */
			snote(1, NOTEXIST, datei);
		else
			snote(1, READERR, datei);
	}
	return(antw);
}

WORD load_datei (const UBYTE* name, RINGP t, BOOLEAN verbose, BOOLEAN *cr_lf)
{
	WORD handle, antw;

	Busy_mouse();
	handle = (short) Fopen(name,0);
	if (handle < 0)
	{
		Last_mouse();
		return(handle);
	}
	else
	{
		LONG	size = Fseek(0,handle,2);

		if (size > 0L)			/* ist was 'drin? */
		{
			Fseek(0,handle,0);
			verbose = verbose && ((size>>10)>=transfer_size);
			if (verbose)
			{
				UBYTE	str[40];
				FILENAME	file;

				strcpy(str, strings(LOADSTR));
				file_name(name,file);
				strcat(str,file);
				start_aktion(str,FALSE,size);
			}
			antw = READ(handle, t, verbose, cr_lf);
			Fclose(handle);
			if (verbose) 
				end_aktion();
		}
		else						/* sonst nix tun! */
			antw = 0;
		Last_mouse();
		return(antw);
	}
}

LOCAL VOID READ2(RINGP t)
{
	LONG	lines, h = 128;
	LINEP start, next;
	BOOLEAN	is_h;

	start = t->head.nachf;
	next = start->nachf;
	for (lines=0; !IS_TAIL(start); lines++)
	{
		next->vorg = start;										/* Rckrichtung verbinden */
		start = next;
		NEXT(next);
	}
	t->lines = lines;
	if (lines>1L)
	{
		col_delete(t->head.nachf);
		t->lines = lines-1;
	}

	start = t->head.nachf;
	while (!IS_TAIL(start))
	{
		if (start->info&h)										/* Zeilen zusammenziehen */
		{
			start->info -= h;
			next = start->nachf;
			if (start->len+next->len<=MAX_LINE_LEN)					/* verbinden */
			{
				is_h = (BOOLEAN)(start->info & h);
				col_concate(&start);
				if (is_h) start->info |= h;
				t->lines--;
			}
			else																	/* zu lang */
			{
				UBYTE	c, *str;
				WORD	i, len;
				
				i = start->len;
				str = TEXT(start)+i;											/* letztes Zeichen+1 */
				while (i>0)														/* Bruch suchen */
				{
					c = *(--str);
					if (c==' ') break;
					i--;
				}
				len = start->len-i;
				if (i>0 && len>0)
				{
					if (len+next->len<=MAX_LINE_LEN)						/* rumschieben */
					{
						INSERT(&next,0,len,TEXT(start)+i);
						REALLOC(&start,i,-len);
					}
					else
					{
						col_split(&start,i); NEXT(start); next = start->nachf;
						t->lines++;
						str = TEXT(next);
						i = 1;
						while (i<next->len)
						{
							c = *str++;
							if (c==' ') break;
							i++;
						}
						if (i<next->len && start->len+i<=MAX_LINE_LEN)	/* hochziehen */
						{
							INSERT(&start,start->len,i,TEXT(next));
							REALLOC(&next,0,-i);
						}
					}
				}
				NEXT(start);
			}
		}
		else
			NEXT(start);
	}
}

LOCAL BOOLEAN READ_BLOCK(RINGP t, UBYTE *j, WORD anz, BOOLEAN *cr_lf, BOOLEAN ende)
{
	UBYTE		c, *zeile;
	WORD		col_len, h;
	UBYTE		*ende_ptr;
	LINEP		start = t->tail.vorg;
	BOOLEAN	nl_found = FALSE;
	
	c = '\x0A';
	h = 255;
	ende_ptr = puffer+anz;
	if (*(ende_ptr - 1) == c)						/* gab es Zeilenende? */
		nl_found = TRUE;
	*ende_ptr = c;										/* fr Ende-Erkennung */
	while (TRUE)
	{
		zeile = j;
		while (TRUE)
			if (*j++==c) 
				break;
		col_len = (short)(j - zeile - 1);
		if (col_len>0 && j[-2]=='\x0D')	/* CR vor einem LF lschen */
		{
			*cr_lf = TRUE;
			col_len--;
		}
		while (col_len>h)						/* Zeile zu lang und zerlegen */
		{
			start = start->nachf = new_col_b(zeile,h);
			start->info |= 128;
			zeile += h;
			col_len -= h;
			if (!ist_mem_frei()) return FALSE;
		}
		if (j>ende_ptr)						/* Puffer ist leer */
		{
			if (col_len)
			{
				start = start->nachf = new_col_b(zeile,col_len);
				start->info |= 128;			/* mu spter verbunden werden */
			}
			break;
		}
		else
		{
			start = start->nachf = new_col_b(zeile,col_len);
		}
		if (!ist_mem_frei()) return FALSE;
	}

	/* zum letzten Mal hier und letzte Zeile hatte \n -> Dummyzeile */
	if (ende & nl_found)		
		start = start->nachf = new_col_b("", 0);

	t->tail.vorg = start;	/* Damit es beim nchsten mal klappt */
	return TRUE;
}

LOCAL WORD READ(WORD handle, RINGP t, BOOLEAN verbose, BOOLEAN *cr_lf)
{
	LONG	bytes;
	WORD	anz;
	LINEP	start;
	BOOLEAN	normal_end;

	*cr_lf = FALSE;
	normal_end = FALSE;
	bytes = 0L;
	while (!normal_end)
	{
		anz = (short)Fread(handle,puff_size,puffer);
		if (anz<puff_size)
		{
			if (anz < 0) 
				break;								/* Sofort abbrechen, weil Fehler	*/
			normal_end = TRUE;
			if (anz == 0) 
				break;
		}
		if (verbose)
		{
			bytes += anz;
			do_aktion(bytes);
		}
		if (!READ_BLOCK(t,puffer,anz,cr_lf, normal_end)) break;
	}
	start = t->tail.vorg;
	start->info &= (~128);
	start->nachf = &t->tail;
	READ2(t);
	if (normal_end)
		return 0;
	else					/* Abbruch wg. Speichermangel oder Lesefehler */
		return -1;
}

WORD infoload(const UBYTE *name, LONG *b, LONG *l)
{
	LONG	bytes, lines;
	WORD	col_len, anz, h, handle;
	UBYTE	c, *j, *zeile, *ende_ptr;
	BOOLEAN	ende, cr_lf;

	Busy_mouse();
	handle = (short)Fopen(name,0);
	if (handle < 0)
	{
		Last_mouse();
		return handle;
	}
	cr_lf = FALSE;
	ende = FALSE;
	h = 255;
	c = '\x0A';
	bytes = lines = 0;
	while (!ende)
	{
		j = puffer;
		anz = (short) Fread(handle,puff_size,j);
		if (anz<puff_size)
		{
			if (anz<=0) break;							/* Sofort abbrechen	*/
			ende = TRUE;									/* Normales Ende		*/
		}
		ende_ptr = puffer+anz;
		*ende_ptr = c;										/* fr Ende-Erkennung */
		while (TRUE)
		{
			zeile = j;
			while (TRUE)
				if (*j++==c) break;
			col_len = (short)(j - zeile - 1);
			if (col_len>0 && j[-2]=='\x0D')	/* CR vor einem LF lschen */
			{
				cr_lf = TRUE;
				col_len--;
			}
			while (col_len>h)						/* Zeile zu lang und zerlegen */
			{
				bytes += h;
				lines++;
				zeile += h;
				col_len -= h;
			}
			if (j>ende_ptr)						/* Puffer ist leer */
			{
				bytes += col_len;					/* mu spter verbunden werden */
				break;
			}
			else
			{
				bytes += col_len;
				lines++;
			}
		}
	}
	bytes += lines;
	if (cr_lf) bytes += lines;
	*b = bytes;
	*l = lines;
	Fclose(handle);
	Last_mouse();
	return 0;
}

/*==================================================================*/

LOCAL VOID write_opt(BUFFER_P buffer, UBYTE* format, ...)
{
	UBYTE str[200], help[20];
	va_list ptr;

	va_start(ptr,format);
	str[0] = '#';
	str[1] = *format++;
	str[2] = EOS;
	while(*format!=EOS)
	{
		strcat(str,"@");
		switch(*format++)
		{
			case 'I'	:	itoa(va_arg(ptr,WORD),help,10);
							strcat(str,help);
							break;
			case 'L' :	ltoa(va_arg(ptr,LONG),help,10);
							strcat(str,help);
							break;
			case 'S' :	strcat(str,va_arg(ptr,UBYTE*));
							break;
		}
	}
	_write(buffer, str, (short) strlen(str), FALSE);
	va_end(ptr);
}

LOCAL VOID read_opt(const UBYTE *format, UBYTE *str, ...)
{
	va_list	ptr;
	WORD		*i;
	LONG		*l;
	UBYTE		*s;

	va_start(ptr,str);
	s = strchr(str,'@');
	while(*format!=EOS)
	{
		if (s==NULL)
		{
			switch (*format++)
			{
				case 'I':	i = va_arg(ptr,WORD*);
								*i = 0;
								break;
				case 'L':	l = va_arg(ptr,LONG*);
								*l = 0L;
								break;
				case 'S':	strcpy(va_arg(ptr,UBYTE*),"");
								break;
			}
		}
		else
		{
			str = s+1;
			s = strchr(str,'@');
			if (s!=NULL) *s = EOS;
			switch (*format++)
			{
				case 'I':	i = va_arg(ptr,WORD*);
								*i = atoi(str);
								break;
				case 'L':	l = va_arg(ptr,LONG*);
								*l = atol(str);
								break;
				case 'S':	strcpy(va_arg(ptr,UBYTE*),str);
								break;
			}
		}
	}
	va_end(ptr);
}

LOCAL	BUFFER	buffer;

LOCAL VOID do_filename(WINDP window)
{
	UBYTE	str[256];

	if (window->class==CLASS_CLIP)
	{
		if (window->opened)
			write_opt(&buffer,"LS","-C");
	}
	if (window->class==CLASS_EDIT || window->class==CLASS_PROJEKT)
	{
		TEXTP		t_ptr = get_text(window->link);
		UBYTE		s[20];

		if (!t_ptr->namenlos)
		{
			strcpy(str,t_ptr->filename);
			strcat(str," ");
			ltoa(t_ptr->ypos+1,s,10);
			strcat(str,s);
			strcat(str," ");
			ltoa(BildPos(t_ptr->xpos,t_ptr->cursor_line,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize)+1,s,10);
			strcat(str,s);
			/* 
			 * Als Text geffnete Projekte: 
			 */
			if (window->class == CLASS_EDIT && 
				(match(t_ptr->filename, "*.QPJ") || match(t_ptr->filename, "*.qpj")))
				strcat(str, " -p ");
			if (!window->opened)
				strcat(str," -o");
			write_opt(&buffer,"LS",str);
		}
	}
}

WORD option_save(BOOLEAN interactiv)
{
	if (interactiv)
	{
		PATH	name;

		if (qed_path[0] == EOS)
			get_path(qed_path, 0);
		if (caseSens(qed_path))
			strcpy(qed_fsel.suffix, "*.qed");
		else
			strcpy(qed_fsel.suffix, "*.QED");
		set_fsel_path(qed_path);
		if (!select_file (&qed_fsel,name, strings(SAVESETUPSTR)))
			return 0;
		make_ext(name,"qed");
		strcpy(qed_path,name);
	}
	Busy_mouse();
	init_buffer(&buffer,qed_path,TRUE,0);
	if (buffer.err)
	{
		Last_mouse();
		return(buffer.err);
	}
	else
	{
		LOC_OPTP	lo;
		LONG		y;
		WORD		x, i, b, b1, b2, k;
		PATH		file;
		WORD		used;
		RECT		r;
		UBYTE		str[256];
		FILENAME	name;

		write_opt(&buffer, "VSI", "qed.INF", INFVERSION);

		k = 0;
	/* frei			k += 1 */
	/* frei			k += 2 */
		if (zeilen_nr)	k += 4;
		if (seiten_nr)	k += 8;
		if (epson)		k += 16;
		if (vorschub)	k += 32;
		if (pruef_prn)	k += 64;
		if (s_round)	k += 128;
		if (s_quant)	k += 256;
		if (s_grkl)		k += 512;
	/* frei			k += 1024 */
		if (s_runter)	k += 2048;
		if (overwrite) k += 4096;
		if (s_global)	k += 8192;
		if (save_opt)	k += 16384;
		b = k;

		k = 0;
		if (is_clip_ondisk())	k += 1;
		if (cu_icon)				k += 2;
		if (cu_window) 			k += 4;
	/*	frei          				k += 8; */
		if (wind_cycle)			k += 16;
		if (cu_width)				k += 32;
		if (cu_high)				k += 64;
		if (df_rekursiv)			k += 128;
		if (wish_no_desktop)		k += 256;
		if (gdos_print)			k += 512;
		if (s_wort)					k += 1024;
		if (blinking_cursor) 	k += 2048;
		if (extern_fontselect)	k += 4096;
	/*	frei							k += 8192; */
	/* frei							k += 16384; */
		b2 = k;

		write_opt(&buffer,"XIIIIIII",b,r_modus,b2,cu_mode,cu_width2,cu_high2,transfer_size);
		write_opt(&buffer,"FIIII", winFont.ID, winFont.Size, druckFont.ID, druckFont.Size);
		write_opt(&buffer,"PIIIII",prn_mode, rand_len, seiten_len, seiten_width, druckDevice);
		write_opt(&buffer,"ESSSS", helpprog, error_str, def_prj_path, SE_Makefile);
		write_opt(&buffer,"GSSII", s_str, df_name, umlaut_from, umlaut_to);
		write_opt(&buffer,"RS",r_str);
		for (i=0; i<HIST_ANZ; i++)
		{
			if (s_history[i][0]==EOS) break;
			write_opt(&buffer,"HS",s_history[i]);
		}
		for (i=0; i<loc_opt_anz; i++)
		{
			lo = &loc_opt[i];
			k = 0;
			if (lo->format_by_load) 
				k += 1;
			if (lo->einruecken)		
				k += 2;
			if (lo->tab)				
				k += 4;
			if (lo->umbrechen) 		
				k += 8;
			if (lo->format_by_paste) 
				k += 16;
			if (lo->backup)
				k += 32;
			write_opt(&buffer,"DSIIISSSS",muster_str[i],k,lo->tabsize,lo->lineal_len,
			                           lo->umbruch_str,lo->wort_str, lo->kurzel, lo->backup_ext);
		}
		/* Fenstergre und Iconposition */
		b = -1;
		while(get_windinfo(&b,&i,&r,&b1,&b2,&used))
			write_opt(&buffer,"WIIIIIIII",i,r.x,r.y,r.w,r.h,b1,b2,used);
		/* geladene Dateien */
		get_all_windows(NIL,SRCH_ANY,do_filename);
		/* Makros */
		b = -1;
		while(get_makro_str(&b,str))
			write_opt(&buffer,"MS",str);
		/* Marken */
		for (b=0; b<5; b++)
		{
			get_marke(b,name,file,&y,&x);
			if (file[0]!=EOS)
				write_opt(&buffer,"SISSLI",b,name,file,y,x);
		}
		/* Icons ohne Fenster */
		get_icon_pos(idisk,&b1, &b2);
		write_opt(&buffer,"IIII",3,b1,b2);
		get_icon_pos(iprinter,&b1,&b2);
		write_opt(&buffer,"IIII",2,b1,b2);
		get_icon_pos(0,&b1,&b2);
		write_opt(&buffer,"IIII",0,b1,b2);

		/* SE-Protokol */
		for (b=0; b < 3; b++)
		{
			if (SE_Shells[b].Name[0]!=EOS)
				write_opt(&buffer, "sISI", b, SE_Shells[b].Name, SE_Shells[b].aktiv); 
		}

		/* Autosave */
		write_opt(&buffer,"AIIIIII", as_text, as_text_min, as_text_ask, 
											  as_proj, as_proj_min, as_proj_ask);
											  
		kill_buffer(&buffer);
		Last_mouse();
		return(buffer.err);
	}
}

WORD option_load(VOID)
{
	BOOLEAN	dummy;
	RING	t;

	if (qed_path[0]==EOS)
	{
		PATH	name;

		if (!select_file (&qed_fsel, name, strings(LOADSETUPSTR)))
			return 0;
		make_ext(name, "qed");
		strcpy(qed_path, name);
	}
	else
	{
		file_name(qed_path, qed_fsel.name);
	}
	init_textring(&t);
	if (load_datei(qed_path,&t,FALSE,&dummy)==0)
	{
		LOC_OPTP	lo_0 = &loc_opt[0], lo;
		WORD	b, b1, b2, i, l, k, x, y, version, muster_nr;
		BOOLEAN	used, bo1;
		RECT	r;
		LONG	anz, line;
		LINEP	lauf;
		UBYTE	*str, s[256], s1[80], s2[80], s4[25];
		PATH	file, s3;

		Busy_mouse();
		muster_nr = 0;
		anz = 0;
		lauf = FIRST(&t);
		str = TEXT(lauf);
		if (*str++=='#' && *str++=='V')
		{
			read_opt("SI",str,s,&version);
			if ((strcmp(s,"qed.INF") == 0 && (version == INFVERSION)))
			{
				NEXT(lauf);
				anz = t.lines-1;
			}
		}
		if (anz<=0)
			note(1,WRONGINF);

		while ((--anz)>=0)
		{
			if (lauf->len>3 && TEXT(lauf)[0]=='#')
			{
				str = TEXT(lauf)+1;
				switch(*str++)
				{
					case 'A' :
						read_opt("IIIIII", str,
												 &as_text, &as_text_min, &as_text_ask, 
											  	 &as_proj, &as_proj_min, &as_proj_ask);
						break;
					case 'D' :
						read_opt("SIIISSS",str, s, &k, &b, &b1, s1, s2, s3);
						if (muster_nr < MUSTER_ANZ)
						{
							if (muster_nr>0) 
								strcpy(muster_str[muster_nr],s);
							lo = &loc_opt[muster_nr];
							l = 1;
							lo->format_by_load = (k&l);	
							k >>= 1;
							lo->einruecken = (k&l);			
							k >>= 1;
							lo->tab = (k&l);					
							k >>= 1;
							lo->umbrechen = (k&l);			
							k >>= 1;
							lo->format_by_paste = (k&l);
							k >>= 1;
							lo->backup = (k&l);
							lo->tabsize = b;
							lo->lineal_len = b1;
							strcpy(lo->umbruch_str,s1);
							str2set(lo->umbruch_str, lo->umbruch_set);
							strcpy(lo->wort_str,s2);
							str2set(lo->wort_str,lo->wort_set);
							strcpy(lo->kurzel, s3);
							strcpy(lo->backup_ext, s4);
							muster_nr++;
						}
						break;
					case 'E' :	
						read_opt("SSSS",str, helpprog, error_str, def_prj_path, SE_Makefile);
						set_def_prj();
						break;
					case 'F' :
						read_opt("IIII",str, &winFont.ID, &winFont.Size, &druckFont.ID, &druckFont.Size);
						font_change();
						font_edit();
						break;
					case 'G' :
						read_opt("SSII", str, s_str, df_name, &umlaut_from, &umlaut_to);
						break;
					case 'H' :
						read_opt("S",str,s);
						for (i=0; i<HIST_ANZ; i++)
							if (s_history[i][0]==EOS) 
								break;
							if (i<HIST_ANZ)
								strcpy(s_history[i],s);
						break;
					case 'I' :	
						read_opt("III",str,&i,&x,&y);
						switch (i)
						{
							case 0 : 
								i = 0;	/* F-Tasten */
								add_icon_to_desk(0,NULL,x,y);
								break;
							case 1 : 
								i = 5; 
								break;
							case 2 : 
								i = 0;	/* PRINTER */
								iprinter = add_icon_to_desk(IPRINTER,icons[IPRINTER].ob_spec.iconblk->ib_ptext, x, y);
								add_icon(printer_type,iprinter);
								break;
							case 3 : 
								i = 0;	/* DISK */
								idisk = add_icon_to_desk(IDISK,icons[IDISK].ob_spec.iconblk->ib_ptext, x, y);
								add_icon(disk_type,idisk);
								break;
							case 4 : 
								i = 3; 
								break;
							case 5 : 
								i = 2; 
								break;
							case 6 : 
								i = 4; 
								break;
							default: 
								i = 0;
						}
						if (i!=0 && version<102)
							add_windinfo(i,NULL,x,y,0);
						break;
					case 'L' :
						read_opt("S",str,s);
						add_diskinfo(s,TRUE);
						break;
					case 'M' :
						set_makro_str(str+1);
						break;
					case 'P' :
						read_opt("IIIII",str,&prn_mode,&rand_len,&seiten_len,&seiten_width,&druckDevice);
						break;
					case 'R' :
						read_opt("S",str,r_str);
						break;
					case 'S' :
						read_opt("ISSLI",str,&b,s,file,&line,&x);
						set_marke(b,s,file,line,x);
						break;
					case 's' :				/* SE-Protokol */
						read_opt("ISI", str, &b, s, &bo1);
						strcpy(SE_Shells[b].Name, s);
						SE_Shells[b].aktiv = bo1;
						break;
					case 'T' :
						read_opt("I",str,&lo_0->tabsize);
						break;
					case 'U' :
						read_opt("S",str,lo_0->umbruch_str);
						str2set(lo_0->umbruch_str,lo_0->umbruch_set);
						break;
					case 'V' :
						read_opt("S",str,lo_0->wort_str);
						str2set(lo_0->wort_str,lo_0->wort_set);
						break;
					case 'W' :
						read_opt("IIIIIIII",str,&i,&r.x,&r.y,&r.w,&r.h,&b1,&b2,&used);
						add_windinfo(i,&r,b1,b2,used);
						break;
					case 'X' :
						read_opt("IIIIIII",str,&b,&r_modus,&b2,&cu_mode,&cu_width2,&cu_high2,&transfer_size);
						if (transfer_size==0) 
							transfer_size = 100;
						l = 1; 
						k = b2;
						clip_ondisk(k&l);	
						k >>= 1;
						cu_icon	= (k&l);
						k >>= 1;
						cu_window = (k&l);		
						k >>= 1;
					/* frei */
						k >>= 1;
						wind_cycle= (k&l);		
						k >>= 1;
						cu_width  = (k&l);		
						k >>= 1;
						cu_high   = (k&l);		
						k >>= 1;
						df_rekursiv=(k&l);		
						k >>= 1;
						no_desktop= (k&l); 
						wish_no_desktop = no_desktop;	
						k >>= 1;
						gdos_print= (k&l);		
						k >>= 1;
						s_wort	 = (k&l);		
						k >>= 1;
						blinking_cursor = (k&l);
						k >>= 1;
						extern_fontselect = (k&l);
						k >>= 1;
					/* frei */
						k >>= 1;
					/* frei */
						
						k = b;
					/* frei */	
						k >>= 1;
					/* frei */
						k >>= 1;
						zeilen_nr = (k&l);		
						k >>= 1;
						seiten_nr = (k&l);		
						k >>= 1;
						epson		 = (k&l);		
						k >>= 1;
						vorschub	 = (k&l);		
						k >>= 1;
						pruef_prn = (k&l);
						k >>= 1;
						s_round   = (k&l);		
						k >>= 1;
						s_quant   = (k&l);		
						k >>= 1;
						s_grkl    = (k&l);		
						k >>= 1;
					/* frei */
						k >>= 1;
						s_runter  = (k&l);		
						k >>= 1;
						overwrite = (k&l);		
						k >>= 1;
						s_global  = (k&l);		
						k >>= 1;
						save_opt  = (k&l);
						break;
					case 'Z' :
						read_opt("I",str,&lo_0->lineal_len); break;
				}
			}
			NEXT(lauf);
		}
		loc_opt_anz = max(1,muster_nr);
		kill_textring(&t);
		Last_mouse();
		return(0);
	}
	else
	{
		kill_textring(&t);
		return(1);
	}
}
