#include <string.h>
#include <stdlib.h>
#include <nkcc.h>
#include <stdio.h>

#include "global.h"
#include "fontsel.h"
#include "menu.h"
#include "obj.h"
#include "scroll.h"

/***************************************************************************/
/* Objekt-Routinen																			*/
/***************************************************************************/

VOID fill_ptext (OBJECT *tree, WORD obj, CONST UBYTE *s)
{
	TEDINFO *ptedinfo;
	WORD	s_len, len;
	UBYTE	*str;

	ptedinfo = (TEDINFO *) get_obspec(tree, obj);
	str = ptedinfo->te_ptext;
	len = ptedinfo->te_txtlen - 1;
	s_len = min((short) strlen(s), len);
	if (ptedinfo->te_just==TE_LEFT)
	{
		str = COPYB (str, s, s_len);
		while ((--len)>=s_len) 
			*str++ = ' ';
		*str = EOS;
	}
	else
	{
		while ((--len)>=s_len) 
			*str++ = ' ';
		str = COPYB (str, s, s_len);
		*str = EOS;
	}
}

VOID set_fstring(OBJECT *tree, WORD obj, CONST UBYTE *s)
{
	UBYTE	*button;
	WORD	len, i;
	
	button = (UBYTE *) get_obspec(tree, obj);
	len = (short) strlen(s);
	for (i = 0; i <= len; i++)
		button[i] = s[i];
}

VOID get_fstring(OBJECT *tree, WORD obj, UBYTE *s)
{
	UBYTE	*button;
	
	button = (UBYTE *) get_obspec(tree, obj);
	strcpy(s, button);
}

VOID hide_objc(OBJECT *tree, WORD obj, BOOLEAN hide)
{
	tree += obj;
	if (hide) tree->ob_flags |= HIDETREE;
	else tree->ob_flags &= (~HIDETREE);
}

VOID disable_objc(OBJECT *tree, WORD obj, BOOLEAN dis)
{
	tree += obj;
	if (dis) tree->ob_state |= DISABLED;
	else tree->ob_state &= (~DISABLED);
}

VOID set_state(OBJECT *tree, WORD obj, BOOLEAN sel)
{
	tree += obj;
	if (sel) tree->ob_state |= SELECTED;
	else	 tree->ob_state &= (~SELECTED);
}

BOOLEAN get_state(OBJECT *tree, WORD obj)
{
	return ((tree[obj].ob_state&SELECTED)!=0);
}

VOID	draw_Dial(DIALINFO *d, WORD obj, WORD deep)
{
	objc_draw(d->Tree, obj, deep, d->x, d->y, d->w, d->h);
}

VOID	draw_Objc(OBJECT *tree, WORD obj, WORD deep)
{
	objc_draw(tree, obj, deep, tree[0].ob_x, tree[0].ob_y, tree[0].ob_width, tree[0].ob_height);
}

long box_size(RECT *box)
{
	return ((box->w+15)>>4)*2L*box->h*planes;
}

VOID get_box(RECT *box, MFDB *buffer)
{
	MFDB	screen;
	WORD	xy [8];

	screen.fd_addr 	= NULL;							/* Bildschirm */
	buffer->fd_w		= box->w;
	buffer->fd_h		= box->h;
	buffer->fd_wdwidth= (box->w+15)>>4; 			/* Anzahl in intS */
	buffer->fd_stand	= FALSE;
	buffer->fd_nplanes= planes;
	buffer->fd_addr	= Malloc(box_size(box));

	if (buffer->fd_addr!=NULL)
	{
		rect2array (box, xy);
		xy[4] = 0;
		xy[5] = 0;
		xy[6] = box->w-1;
		xy[7] = box->h-1;

		set_clip (FALSE, NULL);
		Hide_mouse ();
		vro_cpyfm (vdi_handle, S_ONLY, xy, &screen, buffer);
		Show_mouse ();
	}
	else
		form_dial (FMD_START, box->x, box->y, box->w, box->h,
									 box->x, box->y, box->w, box->h);
}

VOID put_box(RECT *box, MFDB *buffer)
{
	MFDB	screen;
	WORD	xy [8];

	screen.fd_addr 	= NULL;							/* Bildschirm */
	buffer->fd_w		= box->w;
	buffer->fd_h		= box->h;
	buffer->fd_wdwidth= (box->w+15)>>4; 			/* Anzahl in intS */
	buffer->fd_stand	= FALSE;
	buffer->fd_nplanes= planes;

	if (buffer->fd_addr!=NULL)
	{
		xy[0] = 0;
		xy[1] = 0;
		xy[2] = box->w-1;
		xy[3] = box->h-1;
		rect2array (box, &xy[4]);

		set_clip (FALSE, NULL);
		Hide_mouse ();
		vro_cpyfm (vdi_handle, S_ONLY, xy, buffer, &screen);
		Show_mouse ();
		Mfree(buffer->fd_addr);
	}
	else
		form_dial (FMD_FINISH, 0, 0, 0, 0, box->x, box->y, box->w, box->h);
}

VOID dial_size(OBJECT *tree, WORD obj, RECT *box)
{
	WORD	border;

	objc_rect (tree, obj, box, FALSE);

	border = (WORD)((tree [obj].ob_spec.index>>16) & 0x00FFL);
	if (border & 0x0080) border |= 0xFF00; 			/* Rand negativ */
	if (border < 0)
	{
		box->x += border; 									/* Wegen Rand */
		box->y += border;
		box->w -= 2*border;
		box->h -= 2*border;
	}
	if (tree [obj].ob_state & SHADOWED) 				/* Schatten bercksichtigen */
	{
		box->w += 2*abs(border);
		box->h += 2*abs(border);
	}
	if (tree [obj].ob_state & OUTLINED) 				/* Outlined bercksichtigen */
	{
		if (border >= 0)
			border = 3;
		else
			if (border > -3)
				border += 3;
			else
				border = 0;
		box->x -= border; 									/* Wegen Rand */
		box->y -= border;
		box->w += 2*border;
		box->h += 2*border;
	}
}

BOOLEAN background (OBJECT *tree, WORD obj, BOOLEAN get, MFDB *buffer, RECT *box)
{
	dial_size(tree, obj, box);
	if (rc_intersect (&desk, box))				/* Nur auf sichtbarem Schirm zeichnen */
	{
		if (get)
			get_box(box,buffer);
		else
			put_box(box,buffer);
	}
	return (buffer->fd_addr!=NULL);
} /* background */


WORD	find_xobj(OBJECT *tree, WORD xtype)
{
	WORD	d = 0, xt;
	
	
	while (tree[d].ob_next != 0)
	{
		xt = tree[d].ob_type >> 8;
		if (xt == xtype)
			return d;
		d++;
	}
	return -1;
}

/*********************************************************************/
/* Popup-Routinen																		*/
/*********************************************************************/

BOOLEAN	create_popup(POPUP *p, WORD anz, UBYTE *item, WORD len)
{
	WORD	i;
	UBYTE	*zeile;
	
	p->tree = (OBJECT *) Malloc((anz + 2) * sizeof(OBJECT));
	if (p->tree == NULL)
		return FALSE;

	p->max_item = anz;
	p->item_len = len;

	/* 1. Box */
	p->tree[0].ob_next   	= -1;
	p->tree[0].ob_head   	= 1;
	p->tree[0].ob_tail   	= 1;
	p->tree[0].ob_type   	= G_BOX;
	p->tree[0].ob_flags  	= NONE;
	p->tree[0].ob_state  	= OUTLINED;
	p->tree[0].ob_spec.index= (LONG)0x00021100L;		/* Rahmen, Farbe ... */
	p->tree[0].ob_x	     	= 0;
	p->tree[0].ob_y	     	= 0;
	p->tree[0].ob_width  	= 0x0258; /* 600 x 400 */
	p->tree[0].ob_height 	= 0x0190;
	
	/* 2. Box */
	p->tree[1].ob_next   	= 0;
	p->tree[1].ob_head   	= 2;
	p->tree[1].ob_tail   	= 2;
	p->tree[1].ob_type   	= G_BOX;
	p->tree[1].ob_flags  	= NONE;
	p->tree[1].ob_state  	= SHADOWED|DRAW3D;
	p->tree[1].ob_spec.index= (LONG)0x00FF1100L;		/* Rahmen, Farbe ... */
	p->tree[1].ob_x	     	= 0x0002;
	p->tree[1].ob_y	     	= 0x0001;
	p->tree[1].ob_width  	= p->item_len;
	p->tree[1].ob_height 	= 1;

	/* Speicher fr String holen */
	zeile = (UBYTE *) Malloc((p->item_len + 1) * (short) sizeof (UBYTE));
	if (zeile == NULL)
		return FALSE;
		
	strcpy(zeile, item);
	
	/* 1. Men-Eintrag */
	p->tree[2].ob_next   =  1;
	p->tree[2].ob_head   = -1;
	p->tree[2].ob_tail   = -1;
	p->tree[2].ob_type   = G_STRING|0x1700;
	p->tree[2].ob_flags  = SELECTABLE|LASTOB;
	p->tree[2].ob_state  = DRAW3D;
	p->tree[2].ob_spec.free_string = zeile;
	p->tree[2].ob_x	   = 0;
	p->tree[2].ob_y	   = 0;
	p->tree[2].ob_width  = p->item_len;
	p->tree[2].ob_height = 1;
	
	p->akt_item = 2;
	p->first_item = 2;
	
	for(i = 0; i <= p->akt_item; i++)
		rsrc_obfix(p->tree, i);
	
	return TRUE;
}

BOOLEAN	free_popup(POPUP *p)
{
	WORD	i;
	
	if (p->tree == NULL)
		return FALSE;

	for (i=2; i<=p->akt_item; i++)
		Mfree(p->tree[i].ob_spec.free_string);
	Mfree(p->tree);
	return TRUE;
}

BOOLEAN	append_popup(POPUP *p, UBYTE *item)
{
	WORD	i;
	UBYTE	*zeile;
	
	if (p->tree == NULL)
		return FALSE;
	if (p->akt_item >= p->max_item + 1)
		return FALSE;

	/* zurck auf Zeichenkoordinaten */
	for(i = 0; i <= p->akt_item; i++)
	{
		p->tree[i].ob_y      /= sys_hchar; /*gl_hchar;*/
		p->tree[i].ob_width  /= sys_wchar; /*gl_wchar;*/
		p->tree[i].ob_height /= sys_hchar; /*gl_hchar;*/
	}
		
	/* letztes wird vorletztes */
	p->tree[p->akt_item].ob_next  = p->akt_item + 1;
	p->tree[p->akt_item].ob_flags = SELECTABLE;

	p->akt_item++;

	/* Speicher fr String holen */
	zeile = (UBYTE *) Malloc((p->item_len + 1) * (short) sizeof (UBYTE));
	if (zeile == NULL)
		return FALSE;
		
	strcpy(zeile, item);

	/* neues */
	p->tree[p->akt_item].ob_next	=  1;
	p->tree[p->akt_item].ob_head	= -1;
	p->tree[p->akt_item].ob_tail	= -1;
	p->tree[p->akt_item].ob_type	= G_STRING|0x1700;
	p->tree[p->akt_item].ob_flags	= SELECTABLE|LASTOB;	
	p->tree[p->akt_item].ob_state	= DRAW3D;
	p->tree[p->akt_item].ob_spec.free_string = zeile;
	p->tree[p->akt_item].ob_x		= 0;
	p->tree[p->akt_item].ob_y		= p->akt_item - p->first_item;
	p->tree[p->akt_item].ob_width	= p->item_len;
	p->tree[p->akt_item].ob_height= 1;

	/* 2. Box */
	p->tree[1].ob_tail	= p->akt_item;
	p->tree[1].ob_height	= p->akt_item - 1;

	for(i = 0; i <= p->akt_item; i++)
		rsrc_obfix(p->tree, i);
	return TRUE;
}

/******************************************************************************/
/*	G_USERDEFs																						*/
/******************************************************************************/

#define G_VDIT		25			/* Text, der mit aktuellem Zeichensatz gemalt wird */
#define MAX_VDIT	255

LOCAL	USERBLK	vdiT_blk[MAX_VDIT];
LOCAL WORD		vdiT_count = 0;

/***************************************************************************/
/*	Zeichenroutine	fr VDI-Texte in Dialogen 											*/
/***************************************************************************/

LOCAL WORD cdecl draw_vdiT(PARMBLK *pb)
{
	WORD	d;
	RECT	new;
		
	new.x = pb->pb_xc;
	new.y = pb->pb_yc;
	new.w = new.x + pb->pb_wc -1;
	new.h = new.y + pb->pb_hc -1;
	set_clip(TRUE, &new);
	vst_font(vdi_handle, winFont.ID);
	vst_point(vdi_handle, 14, &d, &d, &d, &d);					/* 14pt passt gerade noch ! */
	v_gtext(vdi_handle, pb->pb_x, pb->pb_y, (UBYTE *)pb->pb_parm);
	vst_point(vdi_handle, winFont.Size, &d, &d, &d, &d);		/* wieder zurcksetzen */
	return (pb->pb_currstate & ~SELECTED);
}

BOOLEAN my_fix(OBJECT *tree)
{
	OBJECT	*ob;
	unsigned	type, xtype;

	ob = tree;
	while (TRUE)
	{
		type	= (UBYTE) ob->ob_type;
		xtype = *(UBYTE *) &ob->ob_type;
		if (type == G_STRING && xtype == G_VDIT)
		{
			if (vdiT_count >= MAX_VDIT)
			{
				form_alert(1,"[3][Zu viele G_VDIT!][PANIC]");
				return FALSE;;	
			}
			else
			{
				ob->ob_type	 = G_USERDEF + (G_VDIT << 8);
				vdiT_blk[vdiT_count].ub_code = draw_vdiT;
				vdiT_blk[vdiT_count].ub_parm = (long)ob->ob_spec.free_string;
				ob->ob_spec.userblk= &vdiT_blk[vdiT_count++];
			}
		}
		if (ob->ob_flags&LASTOB) break;
		ob++;
	}
	return TRUE;
}

