/************************************************************************/
/*																								*/
/* Modul: PRINTER.C																		*/
/*																								*/
/************************************************************************/

#include <stdlib.h>		/* itoa, atoi */
#include <string.h>		/* strlen */
#include <vdi.h>

#include "global.h"
#include "block.h"		/* block_copy */
#include "desktop.h"		/* show_icon */
#include "event.h"		/* abbruch */
#include "fontsel.h"		/* druckFont, select_font */
#include "gdos.h"
#include "icon.h"			/* crt_icon */
#include "obj.h"
#include "qed.h"
#include "rsc.h"
#include "set.h"
#include "text.h"
#include "windows.h"
#include "printer.h"

/*
 * Exportierte Variablen
 */
WORD	iprinter, printer_type;


/****** DEFINES **************************************************************/

/****** TYPES ****************************************************************/

/****** VARIABLES ************************************************************/

LOCAL WORD			xpos, prn_x, prn_y, prn_w, prn_h, line_height;
LOCAL WORD			device, prn_handle;
LOCAL BOOLEAN		dev_has_changed;

/****** FUNCTIONS ************************************************************/

LOCAL VOID	plot_char	(UBYTE c);
LOCAL VOID	plot_str		(UBYTE *adr, WORD rand, BOOLEAN tab, WORD tab_size);
LOCAL VOID	plot_int		(WORD z, WORD stellen);

LOCAL VOID		icon_exist	(WORD icon, SET actions);
LOCAL BOOLEAN	icon_test	(WORD icon, WORD action);
LOCAL WORD		icon_edit	(WORD icon, WORD action);
LOCAL BOOLEAN	icon_drag	(WORD icon, WORD source);
LOCAL VOID		drucken		(UBYTE *name, RINGP t, BOOLEAN tab, WORD tab_size);
LOCAL VOID		gdos_drucken(UBYTE *name, RINGP t, BOOLEAN tab, WORD tab_size);
LOCAL BOOLEAN	prn_check 	(VOID);

/***************************************************************************/

LOCAL VOID icon_exist(WORD icon, SET actions)
{
	setclr(actions);
	setincl(actions,DO_INIT);
	setincl(actions,DO_HELP);
}

LOCAL BOOLEAN icon_test(WORD icon, WORD action)
{
	BOOLEAN	erg;

	switch(action)
	{
		case DO_INIT :	erg = TRUE;	break;
		case DO_HELP :	erg = TRUE;	break;
		default		 :	erg = FALSE;
	}
	return erg;
}

/***************************************************************************/
/* Operation durchfhren																	*/
/***************************************************************************/

LOCAL WORD	icon_edit(WORD icon, WORD action)
{
	switch(action)
	{
		case DO_INIT	:	show_icon(icon); break;
		case DO_HELP	:	note (1, HELPPRIN); break;
	}
	return 1;
}
/***************************************************************************/
/* Auf den Drucker wurde ein Icon gezogen												*/
/***************************************************************************/

LOCAL BOOLEAN icon_drag (WORD icon, WORD source)
{
	return (do_icon(source,DO_PRINT)>0);
}

/***************************************************************************/

LOCAL VOID	set_land(WORD land)
{
	if (device==0)
	{
		Cprnout(27);
		Cprnout('R');
		Cprnout(land);
	}
	else
	{
		Cauxout(27);
		Cauxout('R');
		Cauxout(land);
	}
}

LOCAL VOID plot_char(UBYTE c)
{
	WORD	land;

	if (gdos_print)
	{
		UBYTE	str[2];
		WORD	ext[8];

		str[0] = c;
		str[1] = EOS;
		v_gtext(prn_handle,prn_x,prn_y,str);
		vqt_extent(prn_handle,str,ext);
		prn_x += (ext[2] - ext[0]);
	}
	else
	{
		if (epson)
		{
			land = -1;
			switch(c)
			{
				case 64: /*	 @ */
				case 91: /*	 [ */
				case 92: /*	 \ */
				case 93: /*	 ] */
				case 123:/*	 { */
				case 124:/*	 | */
				case 125:/*	 } */
				case 126:	land = 0;	break;		/*	~	*/

				case 129: c	= 125; land = 2;break;	/*		*/
				case 132: c	= 123; land = 2;break;	/*		*/
				case 142: c	=	91; land = 2;break;	/*		*/
				case 148: c	= 124; land = 2;break;	/*		*/
				case 153: c	=  92; land = 2;break;	/*		*/
				case 154: c	=	93; land = 2;break;	/*		*/
				case 158: c	= 126; land = 2;break;	/*		*/
				case 221: c	=	64; land = 2;break;	/*		*/
			}
			if (land>=0) set_land(land);
		}
		if (c=='\n' || c=='\r') xpos = 0;
		else if (xpos>=seiten_width) return;
		else xpos++;
		if (device==0)
			Cprnout(c);
		else
			Cauxout(c);
	}
}

LOCAL VOID plot_ff(VOID)
{
	if (gdos_print)
	{
		prn_x = 0;
		prn_y = 0;
		v_updwk(prn_handle);
		v_clrwk(prn_handle);
	}
	else
		plot_char('\f');
}

LOCAL VOID plot_nl(VOID)
{
	if (gdos_print)
	{
		prn_x = 0;
		prn_y += line_height;
	}
	else
	{
		plot_char('\r');
		plot_char('\n');
	}
}

LOCAL VOID plot_str(UBYTE *adr, WORD rand, BOOLEAN tab, WORD tab_size)
{
	WORD tabH;

	while	((--rand)>=0) plot_char(' ');
	tabH = 0;
	while	(*adr!='\0')
	{
		if (!tabH) tabH = tab_size;
		if (tab && *adr=='\t')
		{
			while (tabH>0)
			{
				plot_char(' ');
				tabH--;
			}
			tabH = tab_size;
		}
		else
		{
			plot_char(*adr);
			tabH --;
		}
		adr++;
	}
}

LOCAL VOID plot_int(WORD z, WORD stellen)
{
	UBYTE str[12];

	itoa(z,str,10);
	plot_str(str, stellen - (short)strlen(str), FALSE, 0);
}

LOCAL VOID epson_reset(VOID)
{
	Cprnout(0x1B);		/* ESC */
	Cprnout(0x40);		/* @ */
}

VOID blk_drucken(UBYTE *name, TEXTP t_ptr)
{
	RING t;

	block_copy(t_ptr,&t);
	if (gdos_print)
		gdos_drucken(name, &t, t_ptr->loc_opt->tab, t_ptr->loc_opt->tabsize);
	else
		drucken(name, &t, t_ptr->loc_opt->tab, t_ptr->loc_opt->tabsize);
	kill_textring(&t);
}

VOID txt_drucken(UBYTE *name, TEXTP t_ptr)
{
	if (gdos_print)
		gdos_drucken(name, &t_ptr->text, t_ptr->loc_opt->tab, t_ptr->loc_opt->tabsize);
	else
		drucken(name, &t_ptr->text, t_ptr->loc_opt->tab, t_ptr->loc_opt->tabsize);
}

LOCAL BOOLEAN open_printer(VOID)
{
	WORD	work_in[11], work_out[57];
	WORD	i, p_xy[4], dummy;

	if (druckDevice < 0)
		return FALSE;
		
	for (i=0; i<=10; i++) 
		work_in[i] = 1;
	work_in[0] = GdosDevices[druckDevice].devID;					/* Drucker */
	work_in[5] = 0;
	work_in[10] = RC;
	v_opnwk(work_in, &prn_handle, work_out);
	if (prn_handle==0) 
		return FALSE;
	prn_w = work_out[0];
	prn_h = work_out[1];
	p_xy[0] = 0;
	p_xy[1] = 0;
	p_xy[2] = prn_w-1;
	p_xy[3] = prn_h+1;
	vs_clip(prn_handle, TRUE, p_xy);
	vq_extnd (prn_handle, FALSE, work_out);	/* So, wg. Fehler im VDI */
	druckFont.Anz = work_out[10];
	druckFont.Anz += vst_load_fonts(prn_handle,0);
	if (druckFont.Anz == 0) 
		return FALSE;
	if (druckFont.ID == 0 || druckFont.Index == 0)
	{
		for (i = 1; i < 100; i++)
		{
			if (druckFont.ID == vqt_name(vdi_handle, i, druckFont.Name))
			{
				druckFont.Index = i;
				break;
			}
		}	
		if (druckFont.Index < 0) 
			return FALSE;
		vst_font (prn_handle, druckFont.ID);
		druckFont.Size = vst_point(prn_handle, druckFont.Size, &dummy, &dummy, &dummy, &dummy);
	}
	return TRUE;
}

LOCAL VOID close_printer(VOID)
{
	vst_unload_fonts(prn_handle,0);
	v_clswk(prn_handle);
}

LOCAL VOID gdos_drucken(UBYTE *name, RINGP t, BOOLEAN tab, WORD tab_size)
{
	WORD	p_xy[8], dummy, eff[3], seite, lzeile, seiten_len;
	UBYTE	datum[9], str[40];
	long	zeile;
	LINEP	lauf;

	if (!open_printer()) 
		return;
	
	/* Schnittstelle nur bei Druckertreibern berprfen!! */
	if (GdosDevices[druckDevice].devID >= 21 && GdosDevices[druckDevice].devID <= 30)
		if (!prn_check())
		{
			close_printer();
			return;
		}
			
	/* Einige Einstellungen fr Metafile */			
	if (GdosDevices[druckDevice].devID >= 31 && GdosDevices[druckDevice].devID <= 40)
	{
		vm_pagesize(prn_handle, 2100, 2970);			/* Blattgre DIN A4 */
/*		vm_coords(prn_handle, 0, 0, 2480, 3508);*/
		vm_coords(prn_handle, 0, 2480, 3508, 0);
		v_meta_extents(prn_handle, 0, 0, 2480, 3508);
	}


	vst_font(prn_handle, druckFont.ID);
	vst_point(prn_handle, druckFont.Size, &dummy, &dummy, &dummy, &dummy);

	vqt_fontinfo(prn_handle, &dummy, &dummy, p_xy, &dummy, eff);
	line_height = p_xy[0] + p_xy[4] + 1;
	vst_alignment(prn_handle, TA_LEFT , TA_ASCENT, &dummy, &dummy); /* Oberkante */
	vqt_extent(prn_handle,"Hallo",p_xy);
	prn_x = 0; prn_y = 0;

	get_datum(datum);
	seiten_len = prn_h/line_height;
	lzeile = seiten_len;
	seite = 1;
	zeile = 0;

	strcpy(str, strings(PRINTSTR));
	strcat(str,name);
	start_aktion(str,TRUE,t->lines);

	Busy_mouse();
	lauf = FIRST(t);
	while (TRUE)
	{
		if (lzeile==seiten_len)
		{
			lzeile = 0;
			if (zeile!=0) plot_ff();
			if (seiten_nr)
			{
				plot_str(strings(FILESTR), 7, FALSE, 0);
				plot_str(name,1,FALSE,0);
				plot_str(strings(PAGESTR), 8, FALSE, 0);
				plot_int(seite++,0);
				plot_str(strings(DATESTR), 8, FALSE, 0);
				plot_str(datum,0,FALSE,0);
				plot_nl();
				plot_nl();
				lzeile = 2;
			}
		}
		if (abbruch())
			if (note(1,BREAK)==1) break;
		do_aktion(zeile);
		zeile++;
		lzeile++;
		plot_str("",rand_len,FALSE,0);
		if (zeilen_nr) 
			plot_int((short) zeile, 4);
		plot_str(TEXT(lauf),(zeilen_nr)?2:0,tab,tab_size);
		plot_nl();
		NEXT(lauf);
		if (IS_TAIL(lauf)) break;
	}
	v_updwk(prn_handle);
	if (vorschub)
		v_clrwk(prn_handle);
	close_printer();
	end_aktion();
	Last_mouse();
	if (epson)
		epson_reset();
}

LOCAL VOID drucken(UBYTE *name, RINGP t, BOOLEAN tab, WORD tab_size)
{
	long	zeile;
	UBYTE	datum[9], str[40];
	WORD	seite, lzeile;
	LINEP	lauf;

	if (!prn_check())
		return;

	device = ((Setprt(-1)&16)!=0);
	get_datum(datum);
	lzeile = seiten_len;
	seite = 1;
	zeile = 0;
	if (epson)
		epson_reset();
	plot_char('\r');

	strcpy(str,strings(PRINTSTR));
	strcat(str,name);
	start_aktion(str,TRUE,t->lines);

	Busy_mouse();
	lauf = FIRST(t);
	while (TRUE)
	{
		if (abbruch())
			if (note(1,BREAK)==1) break;

		if (seiten_len && lzeile==seiten_len)
		{
			lzeile = 0;
			if (zeile!=0) plot_ff();
			if (seiten_nr)
			{
				plot_str(strings(FILESTR), 7, FALSE, 0); plot_str(name,1,FALSE,0);
				plot_str(strings(PAGESTR), 8, FALSE, 0); plot_int(seite++,0);
				plot_str(strings(DATESTR), 8, FALSE, 0); plot_str(datum,0,FALSE,0);
				plot_nl();
				plot_nl();
				lzeile = 2;
			}
		}
		do_aktion(zeile);
		zeile++;
		lzeile++;
		plot_str("",rand_len,FALSE,0);
		if (zeilen_nr) 
			plot_int((short) zeile,4);
		plot_str(TEXT(lauf),(zeilen_nr)?2:0,tab,tab_size);
		plot_nl();
		NEXT(lauf);
		if (IS_TAIL(lauf)) break;
	}
	if (seiten_len && zeile && vorschub)
		plot_ff();
	if (epson)
		epson_reset();
	end_aktion();
	Last_mouse();
}
/*****************************************************************************/
/* Drucker-TestRoutinen																											*/
/*****************************************************************************/

LOCAL BOOLEAN prn_ready (WORD port)
{
	BOOLEAN ready = TRUE;

	switch (port)
	{
		case 0 : ready = Cprnos ()!=0; break;
		case 1 : ready = Cauxos ()!=0; break;
	}
	return (ready);
} /* prn_ready */


LOCAL BOOLEAN prn_check (VOID)
{
	if (pruef_prn)
	{
		device = ((Setprt(-1)&16)!=0);
		while (!prn_ready (device))
			if (note(1, PRNOTRDY)==2)
				return (FALSE);
	}
	return TRUE;
} /* prn_check */

/*****************************************************************************/

LOCAL VOID set_font(VOID)		/* trgt Font in die Box ein */
{
	UBYTE	str[32];

	if ((strlen(druckFont.Name) == 0) || dev_has_changed)
	{
		strcpy(druckFont.Name, GdosDevices[druckDevice].firstName);
		druckFont.Size = GdosDevices[druckDevice].firstSize;
		druckFont.ID = GdosDevices[druckDevice].firstID;
		dev_has_changed = FALSE;
	}
	itoa(druckFont.Size, str, 10);
	set_ptext(printer, DSIZE, str);
	set_ptext(printer, DFONT, druckFont.Name);
}

LOCAL VOID enable_scan(VOID)
{
	hide_objc(printer, DPGEMDOS, TRUE);
	hide_objc(printer, DPGDOS, TRUE);
	hide_objc(printer, DPSCAN, FALSE);
	draw_Objc(printer, DPARAM, MAX_DEPTH);
}

LOCAL VOID enable_gdos(VOID)
{
	hide_objc(printer, DPGEMDOS, TRUE);
	hide_objc(printer, DPGDOS, FALSE);
	hide_objc(printer, DPSCAN, TRUE);
	printer[DLENGTH].ob_flags &= (~EDITABLE);
	printer[DWIDTH].ob_flags &= (~EDITABLE);
	printer[DRAND].ob_flags &= (~EDITABLE);
}

LOCAL VOID enable_gemdos(VOID)
{
	hide_objc(printer, DPGEMDOS, FALSE);
	hide_objc(printer, DPGDOS, TRUE);
	hide_objc(printer, DPSCAN, TRUE);
	printer[DLENGTH].ob_flags |= EDITABLE;
	printer[DWIDTH].ob_flags |= EDITABLE;
	printer[DRAND].ob_flags |= EDITABLE;
}

BOOLEAN prn_options(UBYTE *name, BOOLEAN block)
{
	UBYTE		s[14];
	WORD		antw, save_dev;
	BOOLEAN	save_gdos_print;
	FONTINFO	save_font;
	DIALINFO	dial;
	POPUP		pop;
	WORD		y, i, pos_x, pos_y;
		
	save_gdos_print = gdos_print;
	save_dev = druckDevice;
	memcpy(&save_font, &druckFont, sizeof(FONTINFO));

	if (gdos_print)
	{
		enable_gdos();
		set_ptext(printer, DMODE, " GDOS");
		scanDevices();
		if (devAnzahl == 0)
		{
			set_ptext(printer, DDEVICE, strings(NODEVSTR));
			disable_objc(printer, DDEVICE, TRUE);
			disable_objc(printer, DFONTSEL, TRUE);						
		}
		else
		{
			if (druckDevice >= devAnzahl)		/* aus den Parametern */
			{
				druckDevice = 0;
				dev_has_changed = TRUE;
			}
			open_printer();					/* aktuelle Fontdaten holen */
			close_printer();
			set_ptext(printer,DDEVICE, GdosDevices[druckDevice].devName);
			set_font();
		}
	}
	else
	{
		set_ptext(printer, DMODE, " GEMDOS");
		itoa(seiten_len,s,10);
		set_ptext(printer, DLENGTH, s);
		itoa(seiten_width,s,10);
		set_ptext(printer, DWIDTH, s);
		itoa(rand_len,s,10);
		set_ptext(printer, DRAND, s);
		set_state(printer, DEPSON, epson);
		enable_gemdos();
	}
	set_state(printer, NUMMERNJ, zeilen_nr);
	set_state(printer, SEITENJ, seiten_nr);
	set_state(printer, DVORSCHUB, vorschub);
	set_state(printer, CHECKPRN, pruef_prn);
	if (!block)
	{
		prn_mode = 0;
		undo_state(printer, DBLOCK, SELECTED);
		do_state(printer, DBLOCK, DISABLED);
		do_state(printer, DTEXT, SELECTED);
	}
	else
	{
		undo_state(printer, DBLOCK, DISABLED);
		do_state(printer, DBLOCK, SELECTED);
		undo_state(printer, DTEXT, SELECTED);
	}

	Arrow_mouse();
	open_dial(printer, FALSE, NULL, &dial);
	dial_draw(&dial);
	do
	{
		antw = dial_do(&dial, 0) & 0x7FFF;	
		switch (antw)
		{
			case DMODE:
				if (gdos)
				{
					objc_offset(printer, DMODE, &pos_x, &pos_y);
					pos_x += printer[DMODE].ob_width / 2;
					pos_y += printer[DMODE].ob_height / 2;
					y = popup_menu(popups, PRNPOP, pos_x, pos_y, 
											POPGEMDOS + gdos_print , FALSE, 1 ,NULL);
					if (y > 0 && gdos_print != (y == POPGDOS))
					{
						BOOLEAN	z_nr, s_nr, block;

						gdos_print = (y == POPGDOS);
						z_nr = get_state(printer, NUMMERNJ);
						s_nr = get_state(printer, SEITENJ);
						block = get_state(printer, DBLOCK);
						if (gdos_print)
						{
							set_ptext(printer, DMODE, " GDOS");
							enable_scan();
							scanDevices();
							enable_gdos();
							if (devAnzahl == 0)
							{
								set_ptext(printer, DDEVICE, strings(NODEVSTR));
								disable_objc(printer, DDEVICE, TRUE);
								disable_objc(printer, DFONTSEL, TRUE);
								set_ptext(printer, DFONT, "");
								set_ptext(printer, DSIZE, "");
							}
							else
							{
								if (druckDevice >= devAnzahl)		/* aus den Parametern */
								{
									druckDevice = 0;
									dev_has_changed = TRUE;
								}
								open_printer();			/* aktuelle Fontdaten holen */
								close_printer();
								set_ptext(printer, DDEVICE, GdosDevices[druckDevice].devName);
								set_font();
							}
						}
						else
						{
							set_ptext(printer, DMODE, " GEMDOS");
							enable_gemdos();
						}
						set_state(printer, NUMMERNJ, z_nr);
						set_state(printer, SEITENJ, s_nr);
						set_state(printer, DTEXT, !block);
						set_state(printer, DBLOCK, block);
						draw_Objc(printer, DPARAM, 8);
					}
				}
				undo_state(printer, antw, SELECTED);
				draw_Objc(printer, antw, 1);
				break;
			case DFONTSEL:
				if (devAnzahl > 0)
				{
					if (open_printer())
						if (select_font(prn_handle, druckDevice, FALSE))
						{
							UBYTE	str[32];
								
							itoa(druckFont.Size, str, 10);
							set_ptext(printer, DSIZE, str);
							set_ptext(printer, DFONT, druckFont.Name);
							draw_Objc(printer, DPARAM, 8);
						}
					close_printer();
				}
				undo_state(printer, antw, SELECTED);
				draw_Objc(printer, antw, 1);
				break;			
			case DDEVICE:		
				if (devAnzahl > 0)
				{
					WORD	len;
					UBYTE	str[30];
					
					objc_offset (printer, DDEVICE, &pos_x, &pos_y);
					pos_x += printer[DDEVICE].ob_width / 2;
					pos_y += printer[DDEVICE].ob_height / 2;

					strcpy(str, GdosDevices[0].devName);
					len = (short) strlen(str);
					len = min(DEV_LEN, len);
					if (len < DEV_LEN)
						for (i = len; i < DEV_LEN; i++)
							strcat(str, " ");
					else
					{
						strncpy(str, GdosDevices[0].devName, len);
						str[len] = EOS;
					}
					create_popup(&pop, devAnzahl, str, DEV_LEN);
					for (i = 1; i < devAnzahl; i++)
					{
						if (GdosDevices[i].devName[0] != EOS)
						{
							len = (short) strlen(GdosDevices[i].devName);
							len = min(DEV_LEN, len);
							strncpy(str, GdosDevices[i].devName, len);
							str[len] = EOS;
							append_popup(&pop, str);
						}						
					}
					dial_fix(pop.tree, FALSE);
					y = popup_menu(pop.tree, 1, pos_x, pos_y, druckDevice + pop.first_item, FALSE, 1 ,NULL);
					y -= pop.first_item;
					if (y >= 0)
					{
						druckDevice = y;
						dev_has_changed = TRUE;
						set_ptext(printer, DDEVICE, GdosDevices[druckDevice].devName);
						set_font();
						draw_Objc(printer, DPARAM, 8);
					}
					free_popup(&pop);
				}
				undo_state(printer, antw, SELECTED);
				draw_Objc(printer, antw, 1);
				break;
			default:
				break;
		}
	}
	while (antw != DOK && antw != DRAUS);
	undo_state(printer, antw, SELECTED);
	dial_end(&dial);
	Last_mouse();
	if (antw == DRAUS)
	{
		gdos_print = save_gdos_print;
		druckDevice = save_dev;
		memcpy(&druckFont, &save_font, sizeof(FONTINFO));
	}
	else
	{
		zeilen_nr = get_state(printer, NUMMERNJ);
		seiten_nr = get_state(printer, SEITENJ);
		vorschub = get_state(printer, DVORSCHUB);
		pruef_prn = get_state(printer, CHECKPRN);
		if (get_state(printer, DTEXT))
			prn_mode = 0;
		else
			prn_mode = 1;
		epson = get_state(printer, DEPSON);

		get_ptext(printer, DLENGTH, s);
		if (*s)
			seiten_len = atoi(s);
		else
			seiten_len = 65;
		get_ptext(printer, DWIDTH, s);
		if (*s)
			seiten_width = atoi(s);
		else
			seiten_width = 80;
		get_ptext(printer, DRAND, s);
		if (*s)
			rand_len = atoi(s);
		else
			rand_len = 2;
	}
	return (antw == DSTART);
}

VOID init_printer(VOID)
{
	druckDevice = 0;
	druckFont.Index = 0;
	druckFont.ID = 0;
	druckFont.Name[0] = EOS;
	druckFont.Size = 10;
	druckFont.Anz = 0;
	dev_has_changed = FALSE;
	printer_type = decl_icon_type(icon_test, icon_edit, icon_exist, icon_drag);
}
