#include <nkcc.h>

#include "global.h"
#include "ausgabe.h"		/* BildPos, InterPos */
#include "block.h"
#include "clipbrd.h"		/* undo... */
#include "edit.h"			/* pos_korr, goto_line */
#include "find.h"
#include "icon.h"
#include "kurzel.h"		/* do_kurzel */
#include "makro.h"		/* end_plaF */
#include "qed.h"
#include "scroll.h"		/* COPYB */
#include "set.h"
#include "text.h"
#include "umbruch.h"		/* umbrechen, absatz */
#include "windows.h"
#include "tasten.h"


typedef enum {UP, DOWN, LEFT, RIGHT} DIRECTION;

LOCAL WORD	save_col[130];		/* Gerettete Zeile fr Undo */
LOCAL WORD  save_len;
LOCAL WORD	cntrl_y_col[130];	/* Gerettete Zeile fr cntrl-y */
LOCAL WORD	cntrl_y_len;

LOCAL VOID text_end(TEXTP t_ptr);
LOCAL VOID text_start(TEXTP t_ptr);
LOCAL VOID shift_up(TEXTP t_ptr);
LOCAL VOID up(TEXTP t_ptr);
LOCAL VOID shift_left(TEXTP t_ptr);
LOCAL VOID shift_right(TEXTP t_ptr);
LOCAL VOID right(TEXTP t_ptr);
LOCAL VOID shift_down(TEXTP t_ptr);
LOCAL VOID down(TEXTP t_ptr);
LOCAL VOID cntrl_left(TEXTP t_ptr);
LOCAL VOID cntrl_right(TEXTP t_ptr);
LOCAL VOID shift_cr(TEXTP t_ptr);
LOCAL VOID TABULATOR(TEXTP t_ptr);
LOCAL VOID cntrl_delete(TEXTP t_ptr);
LOCAL VOID shift_delete(TEXTP t_ptr);
LOCAL VOID cntrl_bs(TEXTP t_ptr);
LOCAL VOID shift_ins(TEXTP t_ptr);
LOCAL VOID cntrl_y(TEXTP t_ptr, BOOLEAN add);
LOCAL VOID delete(TEXTP t_ptr);

VOID get_undo_col(TEXTP t_ptr)
{
	if (!test_col_anders() || t_ptr->ypos!=undo_y)
	{
		undo_y = t_ptr->ypos;
		save_len = t_ptr->cursor_line->len;
		COPYW((UBYTE*)save_col,TEXT(t_ptr->cursor_line),save_len);
	}
	add_undo(COL_ANDERS);
}

VOID do_undo_col(TEXTP t_ptr, WORD undo)
{
	LINEP	undo_col;
	WORD	help[130];
	UBYTE	*str;
	WORD	length;

	if (undo==COL_ANDERS)
	{
		undo_col = get_line(&t_ptr->text,undo_y);
		length = undo_col->len;
		COPYW((UBYTE*)help,TEXT(undo_col),length);

		str = REALLOC(&undo_col,0,save_len-length);
		COPYB(str,(UBYTE*)save_col,save_len);

		COPYW((UBYTE*)save_col,(UBYTE*)help,length);
		save_len = length;

		t_ptr->moved++;
		make_chg(t_ptr->link,LINE_CHANGE,undo_y);
		make_chg(t_ptr->link,POS_CHANGE,0);				/* wg. moved */
		t_ptr->cursor_line = undo_col;
		add_undo(COL_ANDERS);
	}
}

LOCAL BOOLEAN pos_move(TEXTP t_ptr, LONG delta)
{
	LINEP lauf;

	lauf = t_ptr->cursor_line;
	if (!t_ptr->up_down)
	{
		t_ptr->desire_x = BildPos(t_ptr->xpos,lauf,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize);
		t_ptr->up_down = TRUE;
	}
	if (delta < 0)	/* rauf */
	{
		if (t_ptr->ypos==0) return FALSE;
		delta = -delta;
		if (delta > t_ptr->ypos) 
			delta = t_ptr->ypos;
		t_ptr->ypos -= delta;
		while (--delta>=0) VORG(lauf);
		t_ptr->cursor_line = lauf;
		t_ptr->xpos = InterPos(t_ptr->desire_x,lauf,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize);
		return TRUE;
	}
	else if (delta>0)	/* runter */
	{
		LONG rest = t_ptr->text.lines-t_ptr->ypos-1;

		if (rest==0) return FALSE;
		if (delta>rest) delta = (WORD)rest;
		t_ptr->ypos += delta;
		while (--delta>=0)	NEXT(lauf);
		t_ptr->cursor_line = lauf;
		t_ptr->xpos = InterPos(t_ptr->desire_x,lauf,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize);
		return TRUE;
	}
	return FALSE;
}

LOCAL VOID text_start(TEXTP t_ptr)
{
	t_ptr->cursor_line = FIRST(&t_ptr->text);
	t_ptr->xpos = 0;
	t_ptr->ypos = 0;
	make_chg(t_ptr->link,POS_CHANGE,0);
}

LOCAL VOID text_end(TEXTP t_ptr)
{
	t_ptr->cursor_line = LAST(&t_ptr->text);
	t_ptr->xpos = 0;
	t_ptr->ypos = t_ptr->text.lines-1;
	make_chg(t_ptr->link,POS_CHANGE,0);
}

LOCAL VOID shift_up(TEXTP t_ptr)
{
	WINDOWP window = get_window(t_ptr->link);

	if (pos_move(t_ptr, -window->w_hight))
	{
		make_chg(t_ptr->link,MOVE_DOWN,window->w_hight);
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else
		end_play();
}

LOCAL VOID up(TEXTP t_ptr)
{
	LINEP	lauf = t_ptr->cursor_line;
	if (!t_ptr->up_down)
	{
		t_ptr->desire_x = BildPos(t_ptr->xpos,lauf,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize);
		t_ptr->up_down = TRUE;
	}
	if (t_ptr->ypos)
	{
		VORG(lauf);
		t_ptr->cursor_line = lauf;
		t_ptr->xpos = InterPos(t_ptr->desire_x,lauf,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize);
		t_ptr->ypos--;
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else
		end_play();
}

LOCAL VOID shift_down(TEXTP t_ptr)
{
	WINDOWP window = get_window(t_ptr->link);

	if (pos_move(t_ptr, window->w_hight))
	{
		make_chg(t_ptr->link,MOVE_UP,window->w_hight);
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else
		end_play();
}

LOCAL VOID down(TEXTP t_ptr)
{
	LINEP	lauf = t_ptr->cursor_line;

	if (!t_ptr->up_down)
	{
		t_ptr->desire_x = BildPos(t_ptr->xpos,lauf,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize);
		t_ptr->up_down = TRUE;
	}
	if (!IS_LAST(t_ptr->cursor_line))
	{
		NEXT(lauf);
		t_ptr->cursor_line = lauf;
		t_ptr->xpos = InterPos(t_ptr->desire_x,lauf,t_ptr->loc_opt->tab,t_ptr->loc_opt->tabsize);
		t_ptr->ypos++;
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else
		end_play();
}

LOCAL VOID shift_left(TEXTP t_ptr)
{
	t_ptr->xpos = 0;
	make_chg(t_ptr->link,POS_CHANGE,0);
}

VOID left(TEXTP t_ptr)
{
	if (t_ptr->xpos)
	{
		t_ptr->xpos--;
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else if (t_ptr->ypos)
	{
		LINEP	lauf = t_ptr->cursor_line;
		
		VORG(lauf);
		t_ptr->cursor_line = lauf;
		t_ptr->xpos = lauf->len;
		t_ptr->ypos--;
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else
		end_play();
}

LOCAL VOID shift_right(TEXTP t_ptr)
{
	t_ptr->xpos = t_ptr->cursor_line->len;
	make_chg(t_ptr->link,POS_CHANGE,0);
}

LOCAL VOID right(TEXTP t_ptr)
{
	LINEP	lauf = t_ptr->cursor_line;

	if (t_ptr->xpos<lauf->len)
	{
		t_ptr->xpos++;
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else if (!IS_LAST(lauf))
	{
		NEXT(lauf);
		t_ptr->cursor_line = lauf;
		t_ptr->xpos = 0;
		t_ptr->ypos++;
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else
		end_play();
}

LOCAL VOID cntrl_left(TEXTP t_ptr)
{
	WORD	xw;
	LONG 	yw;
	UBYTE	*str;
	LINEP	lauf;

	lauf = t_ptr->cursor_line;
	xw = t_ptr->xpos;
	yw = t_ptr->ypos;
	str = TEXT(lauf)+xw;
	do										/* Ersten Buchstaben finden */
	{
		while(xw==0)
		{
			if (yw==0) return;
			VORG(lauf);
			yw --;
			xw = lauf->len;
			str = TEXT(lauf)+xw;
		}
		xw--;
	}while (!setin(t_ptr->loc_opt->wort_set,*(--str)));
	while(xw>0)							/* Ersten NICHT-Buchstaben suchen */
	{
		xw--;
		if (!setin(t_ptr->loc_opt->wort_set,*(--str)))
		{
			xw++;
			break;
		}
	}
	t_ptr->cursor_line = lauf;
	t_ptr->xpos = xw;
	t_ptr->ypos = yw;
	make_chg(t_ptr->link,POS_CHANGE,0);
}

LOCAL VOID cntrl_right(TEXTP t_ptr)
{
	WORD	xw;
	LONG	yw;
	UBYTE	*str;
	LINEP	lauf;

	lauf = t_ptr->cursor_line;
	xw = t_ptr->xpos;
	yw = t_ptr->ypos;
	str = TEXT(lauf)+xw;
	while (TRUE)						/* erster nicht-wort-Buchstabe */
	{
		if (xw==lauf->len || !setin(t_ptr->loc_opt->wort_set,*str))
			break;
		str++; xw++;
	}
	while (TRUE)						/* erster wort-Buchstabe */
	{
		while (xw==lauf->len)
		{
			if (IS_LAST(lauf))
				goto ende;
			NEXT(lauf);
			yw++;
			xw = 0;
			str = TEXT(lauf);
		}
		if (setin(t_ptr->loc_opt->wort_set,*str))
			break;
		str++; xw++;
	}
ende:
	t_ptr->cursor_line = lauf;
	t_ptr->xpos = xw;
	t_ptr->ypos = yw;
	make_chg(t_ptr->link,POS_CHANGE,0);
}

LOCAL VOID shift_cr(TEXTP t_ptr)
{
	LINEP lauf = t_ptr->cursor_line;
	if (IS_LAST(lauf)) return;
	NEXT(lauf);
	t_ptr->cursor_line = lauf;
	t_ptr->xpos = 0;
	t_ptr->ypos++;
	make_chg(t_ptr->link,POS_CHANGE,0);
}

WORD col_offset(LINEP col)
/* Wieviel WS-Zeichen stehen am Anfang der Zeile */
{
	WORD	pos;
	UBYTE	c, *str;

	pos = 0;
	str = TEXT(col);
	c = *str++;
	while ((c=='\t' || c==' ') && pos<col->len)
	{
		pos++;
		c = *str++;
	}
	return pos;
}

WORD einrucken(LINEP *col)
{
	LINEP	vor_col;
	WORD	length;

	vor_col = (*col)->vorg;
	if (!IS_HEAD(vor_col))
	{
		length = col_offset(vor_col);
		if ((*col)->len+length>MAX_LINE_LEN)
		{
			note(1,TOOLONG);
			return 0;
		}
		INSERT(col, 0, length, TEXT(vor_col));
	}
	else
		length = 0;
	return(length);
}

VOID cr(TEXTP t_ptr)
{
	clr_undo();
	if (t_ptr->block)
		blk_delete(t_ptr);
	t_ptr->moved++;
	col_split(&t_ptr->cursor_line,t_ptr->xpos);
	t_ptr->cursor_line->info |= ABSATZ;
	t_ptr->text.lines++;
	NEXT(t_ptr->cursor_line);
	t_ptr->ypos++;
	if (t_ptr->loc_opt->einruecken)
		t_ptr->xpos = einrucken(&t_ptr->cursor_line);
	else
		t_ptr->xpos = 0;

	make_chg(t_ptr->link,SCROLL_DOWN,t_ptr->ypos);
	make_chg(t_ptr->link,LINE_CHANGE,t_ptr->ypos-1);
	make_chg(t_ptr->link,LINE_CHANGE,t_ptr->ypos);
	make_chg(t_ptr->link,POS_CHANGE,0);
	if (t_ptr->loc_opt->umbrechen) umbruch(t_ptr);
}

VOID bs(TEXTP t_ptr)
{
	if (IS_FIRST(t_ptr->cursor_line) && t_ptr->xpos==0)
	{
		end_play();				/* Makro beenden */
	}
	else
	{
		left(t_ptr);
		delete(t_ptr);
	}
}

LOCAL VOID TABULATOR(TEXTP t_ptr)
{
	if (t_ptr->loc_opt->tab)
		insert(t_ptr, '\t');
	else
	{
		WORD xw, tabsize = t_ptr->loc_opt->tabsize;

		xw = BildPos(t_ptr->xpos,t_ptr->cursor_line,t_ptr->loc_opt->tab,tabsize);
		xw = tabsize-(xw%tabsize);
		while((--xw)>=0) insert(t_ptr, ' ');
	}
}

/* Es darf kein Block gesetzt sein */
VOID swap(TEXTP t_ptr)
{
	UBYTE	c;
	WORD	x;

	x = t_ptr->xpos;
	if (x > 0 && x < t_ptr->cursor_line->len)
	{
		get_undo_col(t_ptr);
		c = TEXT(t_ptr->cursor_line)[x];
		TEXT(t_ptr->cursor_line)[x] = TEXT(t_ptr->cursor_line)[x-1];
		TEXT(t_ptr->cursor_line)[x-1] = c;
		t_ptr->xpos++;
		t_ptr->moved++;
		make_chg(t_ptr->link,LINE_CHANGE,t_ptr->ypos);
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
}

VOID insert(TEXTP t_ptr, UBYTE c)
{
	if (t_ptr->block)
		blk_delete(t_ptr);
	get_undo_col(t_ptr);
	if (overwrite && t_ptr->xpos<t_ptr->cursor_line->len)
	{
		TEXT(t_ptr->cursor_line)[t_ptr->xpos] = c;
	}
	else
	{
		if (t_ptr->cursor_line->len==MAX_LINE_LEN)
		{
			note(1,TOOLONG);
			return;
		}
		*(REALLOC(&t_ptr->cursor_line,t_ptr->xpos,1)) = c;
	}
	t_ptr->xpos++;
	t_ptr->moved++;
	make_chg(t_ptr->link,LINE_CHANGE,t_ptr->ypos);
	make_chg(t_ptr->link,POS_CHANGE,0);
	if (t_ptr->loc_opt->umbrechen) umbruch(t_ptr);
}

LOCAL VOID cntrl_delete(TEXTP t_ptr)
/* Lschen bis zum nchsten Wortanfang */
{
	WORD xpos = t_ptr->xpos;

	blk_mark(t_ptr,0);
	while (setin(t_ptr->loc_opt->wort_set,TEXT(t_ptr->cursor_line)[xpos]) && 
				t_ptr->cursor_line->len>xpos)
		xpos++;
	while (!setin(t_ptr->loc_opt->wort_set,TEXT(t_ptr->cursor_line)[xpos]) && 
				t_ptr->cursor_line->len>xpos)
		xpos++;
	t_ptr->xpos = xpos;
	blk_mark(t_ptr,1);
	blk_delete(t_ptr);
}

LOCAL VOID shift_delete(TEXTP t_ptr)
/* Bis zum Zeilenende lschen */
{
	blk_mark(t_ptr,0);
	t_ptr->xpos = t_ptr->cursor_line->len;
	blk_mark(t_ptr,1);
	blk_delete(t_ptr);
}

LOCAL VOID cntrl_bs(TEXTP t_ptr)
/* Lschen wortweise nach links */
{
	WORD 	xpos = t_ptr->xpos - 1;
	
	blk_mark(t_ptr,0);
	while (setin(t_ptr->loc_opt->wort_set, TEXT(t_ptr->cursor_line)[xpos]) && 
			 xpos > 0)
		xpos--;
	while (!setin(t_ptr->loc_opt->wort_set,TEXT(t_ptr->cursor_line)[xpos]) && 
				xpos > 0)
		xpos--;
	if (xpos == 0 && t_ptr->xpos == 1)		/* erstes Zeichen */
		t_ptr->xpos = 0;
	else
		t_ptr->xpos = xpos + 1;
	blk_mark(t_ptr,1);
	blk_delete(t_ptr);
}

LOCAL VOID shift_ins(TEXTP t_ptr)
{
	LINEP col;

	col = new_col_w((UBYTE*)cntrl_y_col,cntrl_y_len);
	col_insert(t_ptr->cursor_line->vorg,col);
	t_ptr->cursor_line = col;
	t_ptr->xpos = 0;
	t_ptr->text.lines++;
	t_ptr->moved++;
	make_chg(t_ptr->link,SCROLL_DOWN,t_ptr->ypos);
	make_chg(t_ptr->link,LINE_CHANGE,t_ptr->ypos);
	make_chg(t_ptr->link,POS_CHANGE,0);
	clr_undo();
}

LOCAL VOID cntrl_y(TEXTP t_ptr, BOOLEAN add)
{
	LINEP	col = t_ptr->cursor_line;

	cntrl_y_len = col->len;
	COPYW((UBYTE*)cntrl_y_col,TEXT(col),cntrl_y_len);
	t_ptr->xpos = 0;
	blk_mark(t_ptr,0);
	if (IS_LAST(col))
	/* letzte Zeile im Text nur krzen */
	{
		t_ptr->xpos = col->len;
		blk_mark(t_ptr,1);
	}
	else
	{
		NEXT(col);
		t_ptr->cursor_line = col;
		t_ptr->ypos++;
		blk_mark(t_ptr,1);
	}
	blk_cut(t_ptr,add);
}

LOCAL VOID delete(TEXTP t_ptr)
{
	LINEP	col = t_ptr->cursor_line;

	if (t_ptr->xpos < col->len)
	{
		get_undo_col(t_ptr);
		t_ptr->moved++;
		REALLOC(&t_ptr->cursor_line,t_ptr->xpos,-1);
		make_chg(t_ptr->link,LINE_CHANGE,t_ptr->ypos);
		make_chg(t_ptr->link,POS_CHANGE,0);						/* wg. moved */
	}
	else if (!IS_LAST(col))
	{
		if (col->len+col->nachf->len>MAX_LINE_LEN)
		{
			note(1,TOOLONG);
			return;
		}
		clr_undo();
		t_ptr->moved++;
		col_concate(&t_ptr->cursor_line);
		t_ptr->text.lines--;
		make_chg(t_ptr->link,SCROLL_UP,t_ptr->ypos+1);
		make_chg(t_ptr->link,LINE_CHANGE,t_ptr->ypos);
		make_chg(t_ptr->link,POS_CHANGE,0);
	}
	else
		end_play();				/* Makro beenden */
	if (t_ptr->loc_opt->umbrechen)	umbruch(t_ptr);
}


/* Deselektiert Block und stellt Cursor entsprechend <dir> */

LOCAL VOID deselect_block(TEXTP t_ptr, DIRECTION dir)
{
	if (dir == LEFT)
	{
		desire_x = BildPos(t_ptr->x1, t_ptr->cursor_line, t_ptr->loc_opt->tab, t_ptr->loc_opt->tabsize);
		desire_y = t_ptr->z1;
	}
	else if (dir == RIGHT)
	{
		desire_x = BildPos(t_ptr->x2, t_ptr->cursor_line, t_ptr->loc_opt->tab, t_ptr->loc_opt->tabsize);
		desire_y = t_ptr->z2;
	}
	blk_demark(t_ptr);
	Icon_edit(t_ptr->link, DO_GOTO);
	t_ptr->up_down = FALSE;
}

BOOLEAN key(TEXTP t_ptr, WINDP window, MKINFO *mk)
{
	WORD		key_code;

#if 0
	/* Alpha-Spezialfunktionen, versteckt auf Shift-Ctrl-F?? */
	if (mk->kreturn == (NKF_FUNC|NKF_CTRL|NKF_SHIFT|NK_F1))
	{
	}
#endif

	/* Sonderbehandlung fr ^Y */
	if (mk->ascii_code == 'Y' && (mk->kreturn & NKF_CTRL))
	{
		if (!t_ptr->block) 
			pos_korr(window, t_ptr, FALSE);
		t_ptr->blk_mark_mode = FALSE;
		blk_demark(t_ptr);
		cntrl_y(t_ptr, mk->shift);
		t_ptr->up_down = FALSE;
		return TRUE;
	}

	if (mk->kreturn & NKF_FUNC)
	{
		if (mk->kreturn & NKF_SHIFT)							/* alle Shift-Codes */
		{
			key_code = mk->kreturn & ~(NKF_FUNC|NKF_SHIFT);
			switch (key_code)
			{
				case NK_RET:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					blk_demark(t_ptr);
					shift_cr(t_ptr);
					t_ptr->up_down = FALSE;
					break;
				case NK_CLRHOME:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode)
						blk_demark(t_ptr);
					text_end(t_ptr);
					t_ptr->up_down = FALSE;
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_UP:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode)
						blk_demark(t_ptr);
					shift_up(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_DOWN:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode)
						blk_demark(t_ptr);
					shift_down(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_LEFT:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode && t_ptr->block)
						blk_demark(t_ptr);
					t_ptr->up_down = FALSE;
					shift_left(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_RIGHT:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode && t_ptr->block)
						blk_demark(t_ptr);
					t_ptr->up_down = FALSE;
					shift_right(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_DEL:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					t_ptr->up_down = FALSE;
					if (t_ptr->block)
						blk_delete(t_ptr);
					else
						shift_delete(t_ptr);
					break;
				default:
					return FALSE;
			} /* switch shift */
		} /* if shift */
		else if (mk->kreturn & NKF_CTRL)					/* alle Control-Codes */
		{
			key_code = mk->kreturn & ~(NKF_FUNC|NKF_CTRL);
			switch (key_code)
			{
				case NK_BS:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					if (t_ptr->block)
						blk_delete(t_ptr);
					else
						cntrl_bs(t_ptr);
					t_ptr->up_down = FALSE;
					break;
				case NK_LEFT:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode && t_ptr->block)
					{
						deselect_block(t_ptr, LEFT);
						break;
					}
					t_ptr->up_down = FALSE;
					cntrl_left(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_RIGHT:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode && t_ptr->block)
					{
						deselect_block(t_ptr, RIGHT);
						break;
					}
					t_ptr->up_down = FALSE;
					cntrl_right(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_DEL:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					t_ptr->up_down = FALSE;
					if (t_ptr->block)
						blk_delete(t_ptr);
					else
						cntrl_delete(t_ptr);
					break;
				default:
					return FALSE;
			} /* switch ctrl */
		} /* if ctrl */
		else if (mk->kreturn & NKF_ALT)						/* alle Alt-Codes */
		{
			key_code = mk->kreturn & ~(NKF_FUNC|NKF_ALT);
			switch (key_code)
			{
				default:
					return FALSE;
			} /* switch alt */
		} /* if alt */
		else 										/* alles restliche Funktionscodes */
		{
			key_code = mk->kreturn & ~NKF_FUNC;
			switch (key_code)
			{
				case NK_ESC :
					if (krz_loaded)
					{
						if (!t_ptr->block) 
							pos_korr(window, t_ptr, FALSE);
						t_ptr->blk_mark_mode = FALSE;
						blk_demark(t_ptr);
						do_kurzel(t_ptr, FALSE);
					}
					break;
				case NK_BS:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					if (t_ptr->block)
						blk_delete(t_ptr);
					else 
						bs(t_ptr);
					t_ptr->up_down = FALSE;
					break;
				case NK_TAB:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					TABULATOR(t_ptr);
					t_ptr->up_down = FALSE;
					break;
				case NK_RET:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					cr(t_ptr);
					t_ptr->up_down = FALSE;
					break;
				case NK_CLRHOME:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode)
						blk_demark(t_ptr);
					text_start(t_ptr);
					t_ptr->up_down = FALSE;
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_UNDO:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					Icon_edit(t_ptr->link, DO_UNDO);
					break;	
				case NK_UP:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode)
						blk_demark(t_ptr);
					up(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_DOWN:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode)
						blk_demark(t_ptr);
					down(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_LEFT:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode && t_ptr->block)
					{
						deselect_block(t_ptr, LEFT);
						break;
					}
					t_ptr->up_down = FALSE;
					left(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_RIGHT:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					if (!t_ptr->blk_mark_mode && t_ptr->block)
					{
						deselect_block(t_ptr, RIGHT);
						break;
					}
					t_ptr->up_down = FALSE;
					right(t_ptr);
					if (t_ptr->blk_mark_mode)
						blk_mark(t_ptr,1);
					break;
				case NK_DEL:
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					t_ptr->up_down = FALSE;
					if (t_ptr->block)
						blk_delete(t_ptr);
					else
						delete(t_ptr);
					break;
				case (NK_ENTER|NKF_NUM) :
					if (!t_ptr->block) 
						pos_korr(window, t_ptr, FALSE);
					t_ptr->blk_mark_mode = FALSE;
					insert(t_ptr, '\r');
					break;
				default:
					return FALSE;
			} /* switch func */
		}
	} /* if NKF_FUNC */
	else													/* keine Funktionstaste */
	{
			if (mk->ascii_code)
			/* Kein Cntrl und ascii nicht Null */
			{
				if (!t_ptr->block) 
					pos_korr(window, t_ptr, FALSE);
				t_ptr->blk_mark_mode = FALSE;
				insert(t_ptr, mk->ascii_code);
				if (krz_loaded) 
					do_kurzel(t_ptr, TRUE);
			}
			else
				return FALSE;
			t_ptr->up_down = FALSE;			/* hier wg. Code-Optimierung */
	}
	return TRUE;
}
