/***************************************************************************/
/*																									*/
/* Modul: TRASH.C																				*/
/*																									*/
/***************************************************************************/
#include <string.h>

#include "global.h"
#include "windows.h"
#include "qed.h"
#include "text.h"		/* clear_text, destruct_text, new_text, text_bytes */
#include "edit.h"
#include "desktop.h"
#include "rsc.h"
#include "icon.h"		/* crt_icon */
#include "trash.h"
#include "set.h"

/*
 * Exportierte Variablen
 */
WORD	itrash, trash_type;


/****** DEFINES ************************************************************/

#define KIND	(NAME|CLOSER|FULLER|MOVER|SIZER|UPARROW|DNARROW|VSLIDE|LFARROW|RTARROW|HSLIDE)
#define FLAGS	(WI_TEXT|WI_FONTSIZE)
#define XFAC	gl_wchar							/* X-Faktor */
#define YFAC	gl_hchar							/* Y-Faktor */
#define INITX	sys_wchar						/* X-Anfangsposition */
#define INITY	42									/* Y-Anfangsposition */

#define XRASTER 8									/* Auf Bytegrenzen posiotionieren */

/****** TYPES **************************************************************/

/****** VARIABLES **********************************************************/

LOCAL TEXTP	trash_text;

/****** FUNCTIONS **********************************************************/

LOCAL BOOLEAN	wi_key		(WINDOWP window, MKINFO *mk);

LOCAL VOID		icon_exist	(WORD icon, SET actions);
LOCAL BOOLEAN	icon_test	(WORD icon, WORD action);
LOCAL WORD		icon_edit	(WORD icon, WORD action);
LOCAL BOOLEAN	icon_drag	(WORD icon, WORD source);

LOCAL VOID		crt_trash	(WORD icon, WINDOWP window);
LOCAL BOOLEAN	open_trash	(WORD icon);

/***************************************************************************/

LOCAL VOID icon_exist(WORD icon, SET actions)
{
	setclr(actions);
/*
	setincl(actions,DO_CLEAR);
	setincl(actions,DO_CLOSE);
	setincl(actions,DO_OPEN);
	setincl(actions,DO_INFO);
*/
	setincl(actions,DO_DESTRUCT);
	setincl(actions,DO_HELP);
	setincl(actions,DO_INIT);
	setincl(actions,DO_REINIT);
}

LOCAL BOOLEAN icon_test(WORD icon, WORD action)
{
	WINDP		window;
	BOOLEAN	erg;

	window = get_window(icon);
	switch (action)
	{
		case DO_CLEAR	:
		case DO_INFO	:		
		case DO_CLOSE	:	erg = window->opened; break;
		case DO_OPEN	:	erg = global_shift; break;
		case DO_DESTRUCT:	erg = TRUE; break;
		case DO_HELP	:	erg = TRUE;	break;
		case DO_INIT	:	erg = TRUE; break;
		case DO_REINIT	:	erg = TRUE;	break;
		default			:	erg = FALSE;
	}
	return erg;
}

LOCAL WORD icon_edit(WORD icon, WORD action)
{
	WORD		erg = 1;
	WINDOWP	window;

	window = get_window(icon);
	switch (action)
	{
		case DO_INIT	:	trash_text = new_text(icon);
								set_text_name(trash_text, icons[ITRASH].ob_spec.iconblk->ib_ptext, TRUE);
								create_window (KIND, CLASS_TRASH,icon,crt_trash);
								break;
		case DO_CLEAR  :  clear_text(trash_text);
								trash_text->cursor = FALSE;
								make_chg(itrash,TOTAL_CHANGE,0);
								restore_edit();
								break;
		case DO_CLOSE	:	close_window(window);
								break;
		case DO_DESTRUCT:	destruct_window(window);
								destruct_text(trash_text);
								trash_text = NULL;
								break;
		case DO_OPEN	:	if (!open_trash(icon))
									erg = -1;
								break;
		case DO_INFO	:	info_edit(icon); break;
		case DO_HELP	:	note (1, HELPTRAS); break;
		case DO_REINIT :	trash_text = new_text(itrash);
								set_text_name(trash_text, icons[ITRASH].ob_spec.iconblk->ib_ptext, TRUE);
								create_window (KIND, CLASS_TRASH,itrash,crt_trash);
								break;
	}
	return erg;
}

/***************************************************************************/
/* Auf den Papierkorb wurde ein Icon gezogen 										*/
/***************************************************************************/

LOCAL BOOLEAN	icon_drag (WORD icon, WORD source)
{
	return (do_icon(source,DO_CLEAR)!=0);
}

/***************************************************************************/

VOID trash_takes_text(RINGPTR r)
{
	TEXTP t_ptr = trash_text;
	
	kill_textring(&t_ptr->text);
	t_ptr->text = *r;
	FIRST(r)->vorg = &t_ptr->text.head;
	LAST(r)->nachf = &t_ptr->text.tail;
	t_ptr->cursor_line = FIRST(&t_ptr->text);
	make_chg(itrash,TOTAL_CHANGE,0);
}

/***************************************************************************/

LOCAL BOOLEAN wi_key(WINDOWP window, MKINFO *mk)
{
	return FALSE;
}

/***************************************************************************/
/* Kreieren eines Fensters																	*/
/***************************************************************************/

LOCAL VOID crt_trash (WORD icon, WINDOWP window)
{
	WORD		initw, inith;

	trash_text->cursor = FALSE;

	initw  = min (desk.w/XFAC*XFAC-7*XFAC, 80*XFAC);
	inith  = desk.h/YFAC*YFAC-7*YFAC;

	window->flags		= FLAGS;
	window->doc.w		= MAX_LINE_LEN;
	window->doc.h		= trash_text->text.lines;
	window->xfac		= XFAC;
	window->yfac		= YFAC;
	window->w_width	= initw/XFAC;
	window->w_hight	= inith/YFAC;
	window->scroll.x	= INITX;
	window->scroll.y	= INITY;
	window->scroll.w	= initw;
	window->scroll.h	= inith;
	window->work		= window->scroll;
	window->draw		= wi_draw_edit;	/* gleiche Routinen wie fr EDIT */
	window->key			= wi_key;
	set_wname(window,icons[ITRASH].ob_spec.iconblk->ib_ptext);
} /* crt_trash */

/***************************************************************************/
/* ffnen des Objekts																		*/
/***************************************************************************/

LOCAL BOOLEAN open_trash (WORD icon)
{
	BOOLEAN	ok;
	WINDOWP	window;

	window = get_window (icon);
	if (window==NULL) return FALSE;

	if (window->opened)
	/* Doppelklick auf Icon und Fenster ist schon offen */
	{
		top_window (window);
		ok = TRUE;
	}
	else
	{
		ok = open_window(window);
	}
	return (ok);
} /* open_trash */

/***************************************************************************/

VOID init_trash (VOID)
{
	trash_type = decl_icon_type(icon_test,icon_edit,icon_exist,icon_drag);
	itrash = add_icon_to_desk(ITRASH,icons[ITRASH].ob_spec.iconblk->ib_ptext, -1, -1);
	add_icon(trash_type,itrash);
}
